(****** Proof for Soundness Lemmas of CCAP: Progress and Preservation ******)
(******                created February 29 2004, revised June 29 2004 ******)

Require Arith.
Require Export ccap.

Lemma CT_CH_lookup_Aux : (t:tspec; psi:codeheapty; ch:codeheap)
 (f:lab; p:assert; g:guar)
 (WF_CH_Aux t psi ch)
  -> (lookupCT psi f p g)
  -> (EX I:iseq | (lookupC ch f I)).
Proof.
Induction psi; Intros.
Inversion H0.
Inversion H1.
Inversion_clear H0.
Rewrite H2; Exists b; Apply lookupC_eq.
Inversion_clear H0.
Cut (EX I:iseq | (lookupC ch0 f I)).
Intro EXI; Inversion_clear EXI.
Exists x; Apply lookupC_neq; Auto.
Apply H with p g0; Auto.
Save.

Lemma CT_CH_lookup : (t:tspec; ch:codeheap)
 (f:lab; p:assert; g:guar; I:iseq)
 (WF_CH t ch)
  -> (lookupCT (tgetCT t) f p g)
  -> (EX I:iseq | (lookupC ch f I)).
Proof.
Intros.
Inversion_clear H.
Apply CT_CH_lookup_Aux with t (tgetCT t) p g; Auto.
Save.

Lemma le_or_gt : (n,m:nat) (le n m) \/ (gt n m).
Proof.
Double Induction 1 2; Auto.
Left; Auto with arith.
Right; Auto with arith.
Intros.
Cut (le n1 n0)\/(gt n1 n0); Auto.
Intro OR; Inversion_clear OR; [Left | Right]; Auto with arith.
Save.

Lemma Progress : (psi1,psi2:codeheapty; inv,p1,p2:assert; A1,G1,A2,G2,g:guar)
 (s:state; ch1,ch2:codeheap; I1,I2:iseq; i:nat)
 (WFW (pairT (pairT inv (pairT psi1 psi2)) (pairT (pairT A1 G1) (pairT A2 G2)))
      p1 p2 g
      ((s, (ch1, ch2)), ((I1, I2), i)))
  -> (inv s)
  /\ (EX W:prog | (Step ((s, (ch1, ch2)), ((I1, I2), i)) W)).
Proof.
Intros.
Inversion_clear H.
(* thread 1 *)
Clear H3; Clear H5; Clear H6; Clear H8.
SplitAllHyps.
Cut inv==inv0; [Intro | Auto].
Rewrite <- H0 in H.
Clear H0; Clear H1; Clear H6; Clear inv0; Clear A0; Clear G0.
Clear p2; Clear i.
Constructor; Auto.
NewInduction I1.
(** seq **)
Exists (((Next c s), (ch1, ch2)), ((I1, I2), (1))); Apply S1_seq.
(** bgt **)
Clear IHI1; NewInduction s.
Inversion_clear H7; SplitAllHyps.
Rewrite <- H4 in H7.
Cut (psi==psi1);
 [Intro eq1; Rewrite eq1 in H1; Clear eq1 H7; Clear psi | Auto].
Cut (le (lookupR b r) (lookupR b r0))
	\/ (gt (lookupR b r) (lookupR b r0)); [Intro OR | Apply le_or_gt].
Inversion_clear OR.
Exists (((a,b),(ch1,ch2)),((I1,I2),(1))); Apply S1_bgtF; Auto.
Cut (EX I:iseq | (lookupC ch1 n I));
 [Intro EXI; Inversion_clear EXI | Apply CT_CH_lookup with
 (pairT (pairT inv psi1) (pairT A1 G1)) p' g'; Auto].
Exists (((a,b),(ch1,ch2)),((x,I2),(1))); Apply S1_bgtT; Auto.
Replace (pairT (pairT inv psi1) (pairT A1 G1)) with t1; Auto.
(** be **)
Clear IHI1; NewInduction s.
Inversion_clear H7; SplitAllHyps.
Rewrite <- H4 in H7.
Cut (psi==psi1);
 [Intro eq1; Rewrite eq1 in H1; Clear eq1 H7; Clear psi | Auto].
Compare (lookupR b r) (lookupR b r0); Intro.
Cut (EX I:iseq | (lookupC ch1 n I));
 [Intro EXI; Inversion_clear EXI | Apply CT_CH_lookup with
 (pairT (pairT inv psi1) (pairT A1 G1)) p' g'; Auto].
Exists (((a,b),(ch1,ch2)),((x,I2),(1))); Apply S1_beT; Auto.
Replace (pairT (pairT inv psi1) (pairT A1 G1)) with t1; Auto.
Exists (((a,b),(ch1,ch2)),((I1,I2),(1))); Apply S1_beF; Auto.
(** ld **)
Clear IHI1; NewInduction s.
Inversion_clear H7; SplitAllHyps.
Rewrite <- H4 in H0.
Cut (inv0==inv);
 [Intro eqinv; Rewrite eqinv in H1; Clear H0 eqinv; Clear inv0 | Auto].
Cut (EX n0:nat |
             (hread a (plus n (lookupR b r0)))=(Some nat n0)
             /\(inv (a,(updateR b r n0)))
               /\(p' (a,(updateR b r n0)))
                 /\((s:(heap*rfile))
                     (g' (a,(updateR b r n0)) s)->(g (a,b) s)));
 [Clear H1; Intro H1 | Auto].
Inversion_clear H1; SplitAllHyps.
Exists (((a,(updateR b r x)),(ch1,ch2)),((I1,I2),(1))); Apply S1_ld; Auto.
(** st **)
Clear IHI1; NewInduction s.
Inversion_clear H7; SplitAllHyps.
Rewrite <- H4 in H0.
Cut (inv0==inv);
 [Intro eqinv; Rewrite eqinv in H1; Clear H0 eqinv; Clear inv0 | Auto].
Exists ((((hwrite a (plus n (lookupR b r)) (lookupR b r0)),b),(ch1,ch2)),((I1,I2),(1))).
Cut (hindom (plus n (lookupR b r)) a)
          /\(H1:heap) ((hwrite a (plus n (lookupR b r)) (lookupR b r0))=H1
             ->((inv (H1,b))/\(p' (H1,b)))
               /\((s:(heap*rfile))(g' (H1,b) s)->(g (a,b) s)));
 [Clear H1; Intro H1; SplitAllHyps | Auto].
Apply S1_st; Auto.
(** yield **)
Exists ((s,(ch1,ch2)),((I1,I2),(1))); Apply S1_yld; Auto.
(* jd *)
Inversion_clear H7.
Rewrite <- H4 in H0; Inversion_clear H0.
Cut (EX I:iseq | (lookupC ch1 n I));
 [Intro | Apply CT_CH_lookup with
   (pairT (pairT inv psi1) (pairT A1 G1)) p' g'; Auto].
Inversion_clear H0.
Exists ((s,(ch1,ch2)),((x,I2),(1))); Apply S1_jd; Auto.
Replace (pairT (pairT inv psi1) (pairT A1 G1)) with t1; Auto.
Replace (tgetCT (pairT (pairT inv psi1) (pairT A1 G1))) with psi; Auto.
(* thread 2 *)
Clear H2; Clear H5; Clear H6; Clear H8.
SplitAllHyps.
Cut inv==inv0; [Intro | Auto].
Rewrite <- H0 in H.
Clear H0; Clear H1; Clear H6; Clear inv0; Clear A0; Clear G0.
Clear p1; Clear i.
Constructor; Auto.
NewInduction I2.
(** seq **)
Exists (((Next c s), (ch1, ch2)), ((I1, I2), (2))); Apply S2_seq.
(** bgt **)
Clear IHI2; NewInduction s.
Inversion_clear H7; SplitAllHyps.
Rewrite <- H5 in H7.
Cut (psi==psi2);
 [Intro eq2; Rewrite eq2 in H1; Clear eq2 H7; Clear psi0 | Auto].
Cut (le (lookupR b r) (lookupR b r0))
	\/ (gt (lookupR b r) (lookupR b r0)); [Intro OR | Apply le_or_gt].
Inversion_clear OR.
Exists (((a,b),(ch1,ch2)),((I1,I2),(2))); Apply S2_bgtF; Auto.
Cut (EX I:iseq | (lookupC ch2 n I));
 [Intro EXI; Inversion_clear EXI | Apply CT_CH_lookup with
 (pairT (pairT inv psi2) (pairT A2 G2)) p' g'; Auto].
Exists (((a,b),(ch1,ch2)),((I1,x),(2))); Apply S2_bgtT; Auto.
Replace  (pairT (pairT inv psi2) (pairT A2 G2)) with t2; Auto.
(** be **)
Clear IHI2; NewInduction s.
Inversion_clear H7; SplitAllHyps.
Rewrite <- H5 in H7.
Cut (psi==psi2);
 [Intro eq2; Rewrite eq2 in H1; Clear eq2 H7; Clear psi | Auto].
Compare (lookupR b r) (lookupR b r0); Intro.
Cut (EX I:iseq | (lookupC ch2 n I));
 [Intro EXI; Inversion_clear EXI | Apply CT_CH_lookup with
 (pairT (pairT inv psi2) (pairT A2 G2)) p' g'; Auto].
Exists (((a,b),(ch1,ch2)),((I1,x),(2))); Apply S2_beT; Auto.
Replace (pairT (pairT inv psi2) (pairT A2 G2)) with t2; Auto.
Exists (((a,b),(ch1,ch2)),((I1,I2),(2))); Apply S2_beF; Auto.
(** ld **)
Clear IHI2; NewInduction s.
Inversion_clear H7; SplitAllHyps.
Rewrite <- H5 in H0.
Cut (inv0==inv);
 [Intro eqinv; Rewrite eqinv in H1; Clear H0 eqinv; Clear inv0 | Auto].
Cut (EX n0:nat |
             (hread a (plus n (lookupR b r0)))=(Some nat n0)
             /\(inv (a,(updateR b r n0)))
               /\(p' (a,(updateR b r n0)))
                 /\((s:(heap*rfile))
                     (g' (a,(updateR b r n0)) s)->(g (a,b) s)));
 [Clear H1; Intro H1 | Auto].
Inversion_clear H1; SplitAllHyps.
Exists (((a,(updateR b r x)),(ch1,ch2)),((I1,I2),(2))); Apply S2_ld; Auto.
(** st **)
Clear IHI2; NewInduction s.
Inversion_clear H7; SplitAllHyps.
Rewrite <- H5 in H0.
Cut (inv0==inv);
 [Intro eqinv; Rewrite eqinv in H1; Clear H0 eqinv; Clear inv0 | Auto].
Exists ((((hwrite a (plus n (lookupR b r)) (lookupR b r0)),b),(ch1,ch2)),((I1,I2),(2))).
Cut (hindom (plus n (lookupR b r)) a)
          /\(H1:heap) ((hwrite a (plus n (lookupR b r)) (lookupR b r0))=H1
             ->((inv (H1,b))/\(p' (H1,b)))
               /\((s:(heap*rfile))(g' (H1,b) s)->(g (a,b) s)));
 [Clear H3; Intro H3; SplitAllHyps | Auto].
Apply S2_st; Auto.
(** yield **)
Exists ((s,(ch1,ch2)),((I1,I2),(2))); Apply S2_yld; Auto.
(* jd *)
Inversion_clear H7.
Rewrite <- H5 in H0; Inversion_clear H0.
Cut (EX I:iseq | (lookupC ch2 n I));
 [Intro | Apply CT_CH_lookup with
   (pairT (pairT inv psi2) (pairT A2 G2)) p' g'; Auto].
Inversion_clear H0.
Exists ((s,(ch1,ch2)),((I1,x),(2))); Apply S2_jd; Auto.
Replace (pairT (pairT inv psi2) (pairT A2 G2)) with t2; Auto.
Replace (tgetCT (pairT (pairT inv psi2) (pairT A2 G2))) with psi; Auto.
Save.

Lemma CT_CH_Infer_Aux : (t:tspec; psi:codeheapty; ch:codeheap)
 (f:lab; p:assert; g:guar; I:iseq)
 (WF_CH_Aux t psi ch)
  -> (lookupCT psi f p g)
  -> (lookupC ch f I)
  -> (Infer t p g I).
Proof.
Intros.
Induction H.
Inversion H0.
Compare f0 f; Intro.
Rewrite e in H0; Rewrite e in H1.
Inversion H0; [Idtac | Absurd f=f; Auto].
Inversion H1; [Idtac | Absurd f=f; Auto].
Rewrite <- H8; Rewrite <- H9; Rewrite <- H14; Assumption.
Apply HrecH.
Inversion H0; [Absurd f=f; Auto | Auto].
Inversion H1; [Absurd f=f; Auto | Auto].
Save.

Lemma CT_CH_Infer : (t:tspec; ch:codeheap; f:lab; p:assert; g:guar; I:iseq)
 (WF_CH t ch)
  -> (lookupCT (tgetCT t) f p g)
  -> (lookupC ch f I)
  -> (Infer t p g I).
Proof.
Intros.
Inversion H; Apply CT_CH_Infer_Aux with (tgetCT t) ch f; Auto.
Save.

Lemma Preservation : (i:nat; psi1,psi2:codeheapty; inv,p1,p2:assert)
 (A1,G1,A2,G2,g:guar; s:state; ch1,ch2:codeheap; I1,I2:iseq; W:prog)
 (WFW (pairT (pairT inv (pairT psi1 psi2)) (pairT (pairT A1 G1) (pairT A2 G2)))
      p1 p2 g
      ((s, (ch1, ch2)), ((I1, I2), i)))
  -> ((s,s':state) (G1 s s') -> (A2 s s'))
  -> ((s,s':state) (G2 s s') -> (A1 s s'))
  -> (Step ((s, (ch1, ch2)), ((I1, I2), i)) W)
  -> (EXT p1':assert | (EXT p2':assert | (EXT g':guar |
	(WFW (pairT (pairT inv (pairT psi1 psi2))
	    (pairT (pairT A1 G1) (pairT A2 G2)))
	     p1' p2' g' W)))).
Proof.
Intros.
Inversion H; Clear H; SplitAllHyps.
(* thread 1 *)
Clear H12 H13 H14 H15 H3 H4 H5 H6 H7 phi p0 p3 g0.
Rewrite <- H8 in H2; Clear H8 i.
Cut (inv==inv0); [Clear H10 ; Intro | Auto].
Rewrite <- H3 in H; Rewrite <- H3 in H19; Clear H3 inv0.
Cut (pairT A2 G2)==(pairT A0 G0); [Clear H24; Intro | Auto].
Inversion H3; Clear H3.
Rewrite <- H5 in H19; Rewrite <- H6 in H21; Rewrite <- H5; Rewrite <- H6;
  Clear H5 H6 A0 G0.
Clear h0 ch0 ch3 b1 b2.
NewInduction I1.
(** seq **)
Clear IHI1.
Inversion_clear H20; Inversion_clear H2.
Exists p'; Exists p2; Exists g'.
Rewrite <- H17 in H3.
Cut (inv0==inv); [Intro eqinv; Rewrite eqinv in H4; Clear H3 eqinv | Auto].
Cut (inv (Next c s))/\(p' (Next c s))
  /\((s':(heap*rfile))(g' (Next c s) s')->(g s s'));
 [Intro; SplitAllHyps | Auto].
Apply wf1_w with t1 t2 inv A2 G2; Auto.
(** bgt **)
Clear IHI1.
NewInduction s.
Inversion_clear H20; Inversion_clear H2.
SplitAllHyps.
Clear H4 H5 H2.
Exists p''; Exists p2; Exists g''.
Rewrite <- H17 in H8.
Cut (inv0==inv); [Intro eqinv; Rewrite eqinv in H6; Clear H8 eqinv | Auto].
Cut (p'' (a,b))/\((s:(heap*rfile))(g'' (a,b) s)->(g (a,b) s));
 [Intro; SplitAllHyps | Auto].
Apply wf1_w with t1 t2 inv A2 G2; Auto.
SplitAllHyps.
Clear H6 H7.
Exists p'; Exists p2; Exists g'.
Rewrite <- H17 in H8.
Cut (inv0==inv); [Intro eqinv; Rewrite eqinv in H5; Clear H8 eqinv | Auto].
Rewrite <- H17 in H2.
Cut (psi==psi1); [Intro eqpsi1; Rewrite eqpsi1 in H4; Clear H2 eqpsi1 | Auto].
Cut (p' (a,b))/\((s:(heap*rfile))(g' (a,b) s)->(g (a,b) s));
 [Intro; SplitAllHyps | Auto].
Apply wf1_w with t1 t2 inv A2 G2; Auto.
Rewrite <- H17.
Apply CT_CH_Infer with ch1 n; Auto.
Replace (getT1
       (pairT (pairT inv (pairT psi1 psi2))
         (pairT (pairT A1 G1) (pairT A2 G2)))) with t1; Auto.
(** be **)
Clear IHI1.
NewInduction s.
Inversion_clear H20; Inversion_clear H2.
SplitAllHyps.
Clear H4 H5 H2.
Exists p''; Exists p2; Exists g''.
Rewrite <- H17 in H8.
Cut (inv0==inv); [Intro eqinv; Rewrite eqinv in H6; Clear H8 eqinv | Auto].
Cut (p'' (a,b))/\((s:(heap*rfile))(g'' (a,b) s)->(g (a,b) s));
 [Intro; SplitAllHyps | Auto].
Apply wf1_w with t1 t2 inv A2 G2; Auto.
SplitAllHyps.
Clear H6 H7.
Exists p'; Exists p2; Exists g'.
Rewrite <- H17 in H8.
Cut (inv0==inv); [Intro eqinv; Rewrite eqinv in H5; Clear H8 eqinv | Auto].
Rewrite <- H17 in H2.
Cut (psi==psi1); [Intro eqpsi1; Rewrite eqpsi1 in H4; Clear H2 eqpsi1 | Auto].
Cut (p' (a,b))/\((s:(heap*rfile))(g' (a,b) s)->(g (a,b) s));
 [Intro; SplitAllHyps | Auto].
Apply wf1_w with t1 t2 inv A2 G2; Auto.
Rewrite <- H17.
Apply CT_CH_Infer with ch1 n; Auto.
Replace (getT1
       (pairT (pairT inv (pairT psi1 psi2))
         (pairT (pairT A1 G1) (pairT A2 G2)))) with t1; Auto.
(** ld **)
Clear IHI1.
NewInduction s.
Inversion_clear H20; Inversion_clear H2.
Exists p'; Exists p2; Exists g'.
Rewrite <- H17 in H3.
Cut (inv0==inv); [Intro eqinv; Rewrite eqinv in H4; Clear H3 eqinv | Auto].
Cut (EX n0:nat |
             (hread a (plus n (lookupR b r0)))=(Some nat n0)
             /\(inv (a,(updateR b r n0)))
               /\(p' (a,(updateR b r n0)))
                 /\((s:(heap*rfile))
                     (g' (a,(updateR b r n0)) s)->(g (a,b) s)));
 [Clear H4; Intro H4 | Auto].
Inversion_clear H4; SplitAllHyps.
Rewrite H3 in H7.
Inversion H7.
Replace n0 with x; Auto.
Apply wf1_w with t1 t2 inv A2 G2; Auto.
(** st **)
Clear IHI1.
NewInduction s.
Inversion_clear H20; Inversion_clear H2.
Exists p'; Exists p2; Exists g'.
Rewrite <- H17 in H3.
Cut (inv0==inv); [Intro eqinv; Rewrite eqinv in H4; Clear H3 eqinv | Auto].
Cut (hindom (plus n (lookupR b r)) a)
          /\((H1:heap)
              (hwrite a (plus n (lookupR b r)) (lookupR b r0))=H1
              ->((inv (H1,b))/\(p' (H1,b)))
                /\((s:(heap*rfile))(g' (H1,b) s)->(g (a,b) s)));
 [Clear H4; Intro H4 | Auto].
Inversion_clear H4.
Cut ((inv ((hwrite a (plus n (lookupR b r)) (lookupR b r0)),b))
  /\(p' ((hwrite a (plus n (lookupR b r)) (lookupR b r0)),b)))
  /\((s:(heap*rfile))(g' ((hwrite a (plus n (lookupR b r)) (lookupR b r0)),b) s)->(g (a,b) s)); [Clear H3; Intro H3; Inversion_clear H3 | Auto].
Apply wf1_w with t1 t2 inv A2 G2; Auto.
(** yield **)
Clear IHI1.
Inversion_clear H20; Inversion_clear H2.
Inversion_clear H3.
Rewrite <- H17 in H2.
Cut inv0==inv; [Intro; Clear H2 | Auto].
Rewrite H3 in H4; Rewrite H3 in H5; Clear H3 inv0.
Rewrite <- H17 in H8.
Cut (pairT A G)==(pairT A1 G1); [Clear H8; Intro | Auto].
Inversion H2; Clear H2.
Rewrite H8 in H5; Rewrite H9 in H6; Clear H8 H9 A G W.
Inversion_clear H7.
(*** case 1 ***)
Rewrite H2; Clear H2 i.
Exists p1; Exists p2; Exists G1; Apply wf1_w with t1 t2 inv A2 G2; Auto.
Intros; Apply H19 with s; Auto.
(*** case 2 ***)
Rewrite H2; Clear H2 i.
Exists p1; Exists p2; Exists G2; Apply wf2_w with t1 t2 inv A1 G1; Auto.
Intros; Apply H5 with s; Auto.
(** jd **)
Inversion_clear H20; Inversion_clear H2.
Inversion_clear H3.
Rewrite <- H17 in H7.
Cut inv0==inv; [Intro; Clear H7 | Auto].
Rewrite H3 in H5; Clear H3 inv0.
Rewrite <- H17 in H2.
Cut psi==psi1; [Clear H2; Intro | Auto].
Rewrite H2 in H4; Clear H2 psi W.
Exists p'; Exists p2; Exists g'.
Cut (p' s)/\((s':(heap*rfile))(g' s s')->(g s s'));
 [Clear H5; Intro H5; Inversion_clear H5 | Auto].
Apply wf1_w with t1 t2 inv A2 G2; Auto.
Rewrite <- H17.
Apply CT_CH_Infer with ch1 n; Auto.
Replace (getT1
       (pairT (pairT inv (pairT psi1 psi2))
         (pairT (pairT A1 G1) (pairT A2 G2)))) with t1; Auto.
(* thread 2 *)
Clear H12 H13 H14 H15 H3 H4 H5 H6 H7 phi p0 p3 g0.
Rewrite <- H8 in H2; Clear H8 i.
Cut (inv==inv0); [Clear H10 ; Intro | Auto].
Rewrite <- H3 in H; Rewrite <- H3 in H19; Clear H3 inv0.
Cut (pairT A1 G1)==(pairT A0 G0); [Clear H24; Intro | Auto].
Inversion H3; Clear H3.
Rewrite <- H5 in H19; Rewrite <- H6 in H21; Rewrite <- H5; Rewrite <- H6;
  Clear H5 H6 A0 G0.
Clear h0 ch0 ch3 b1 b2.
NewInduction I2.
(** seq **)
Clear IHI2.
Inversion_clear H20; Inversion_clear H2.
Exists p1; Exists p'; Exists g'.
Rewrite <- H23 in H3.
Cut (inv0==inv); [Intro eqinv; Rewrite eqinv in H4; Clear H3 eqinv | Auto].
Cut (inv (Next c s))/\(p' (Next c s))
  /\((s':(heap*rfile))(g' (Next c s) s')->(g s s'));
 [Intro; SplitAllHyps | Auto].
Apply wf2_w with t1 t2 inv A1 G1; Auto.
(** bgt **)
Clear IHI2.
NewInduction s.
Inversion_clear H20; Inversion_clear H2.
SplitAllHyps.
Clear H4 H5 H2.
Exists p1; Exists p''; Exists g''.
Rewrite <- H23 in H8.
Cut (inv0==inv); [Intro eqinv; Rewrite eqinv in H6; Clear H8 eqinv | Auto].
Cut (p'' (a,b))/\((s:(heap*rfile))(g'' (a,b) s)->(g (a,b) s));
 [Intro; SplitAllHyps | Auto].
Apply wf2_w with t1 t2 inv A1 G1; Auto.
SplitAllHyps.
Clear H6 H7.
Exists p1; Exists p'; Exists g'.
Rewrite <- H23 in H8.
Cut (inv0==inv); [Intro eqinv; Rewrite eqinv in H5; Clear H8 eqinv | Auto].
Rewrite<- H23 in H2.
Cut (psi==psi2); [Intro eqpsi2; Rewrite eqpsi2 in H4; Clear H2 eqpsi2 | Auto].
Cut (p' (a,b))/\((s:(heap*rfile))(g' (a,b) s)->(g (a,b) s));
 [Intro; SplitAllHyps | Auto].
Apply wf2_w with t1 t2 inv A1 G1; Auto.
Rewrite <- H23.
Apply CT_CH_Infer with ch2 n; Auto.
Replace (getT2
       (pairT (pairT inv (pairT psi1 psi2))
         (pairT (pairT A1 G1) (pairT A2 G2)))) with t2; Auto.
(** be **)
Clear IHI2.
NewInduction s.
Inversion_clear H20; Inversion_clear H2.
SplitAllHyps.
Clear H4 H5 H2.
Exists p1; Exists p''; Exists g''.
Rewrite <- H23 in H8.
Cut (inv0==inv); [Intro eqinv; Rewrite eqinv in H6; Clear H8 eqinv | Auto].
Cut (p'' (a,b))/\((s:(heap*rfile))(g'' (a,b) s)->(g (a,b) s));
 [Intro; SplitAllHyps | Auto].
Apply wf2_w with t1 t2 inv A1 G1; Auto.
SplitAllHyps.
Clear H6 H7.
Exists p1; Exists p'; Exists g'.
Rewrite <- H23 in H8.
Cut (inv0==inv); [Intro eqinv; Rewrite eqinv in H5; Clear H8 eqinv | Auto].
Rewrite <- H23 in H2.
Cut (psi==psi2); [Intro eqpsi2; Rewrite eqpsi2 in H4; Clear H2 eqpsi2 | Auto].
Cut (p' (a,b))/\((s:(heap*rfile))(g' (a,b) s)->(g (a,b) s));
 [Intro; SplitAllHyps | Auto].
Apply wf2_w with t1 t2 inv A1 G1; Auto.
Rewrite <- H23.
Apply CT_CH_Infer with ch2 n; Auto.
Replace (getT2
       (pairT (pairT inv (pairT psi1 psi2))
         (pairT (pairT A1 G1) (pairT A2 G2)))) with t2; Auto.
(** ld **)
Clear IHI2.
NewInduction s.
Inversion_clear H20; Inversion_clear H2.
Exists p1; Exists p'; Exists g'.
Rewrite <- H23 in H3.
Cut (inv0==inv); [Intro eqinv; Rewrite eqinv in H4; Clear H3 eqinv | Auto].
Cut (EX n0:nat |
             (hread a (plus n (lookupR b r0)))=(Some nat n0)
             /\(inv (a,(updateR b r n0)))
               /\(p' (a,(updateR b r n0)))
                 /\((s:(heap*rfile))
                     (g' (a,(updateR b r n0)) s)->(g (a,b) s)));
 [Clear H4; Intro H4 | Auto].
Inversion_clear H4; SplitAllHyps.
Rewrite H3 in H7.
Inversion H7.
Replace n0 with x; Auto.
Apply wf2_w with t1 t2 inv A1 G1; Auto.
(** st **)
Clear IHI2.
NewInduction s.
Inversion_clear H20; Inversion_clear H2.
Exists p1; Exists p'; Exists g'.
Rewrite <- H23 in H3.
Cut (inv0==inv); [Intro eqinv; Rewrite eqinv in H4; Clear H3 eqinv | Auto].
Cut (hindom (plus n (lookupR b r)) a)
          /\((H1:heap)
              (hwrite a (plus n (lookupR b r)) (lookupR b r0))=H1
              ->((inv (H1,b))/\(p' (H1,b)))
                /\((s:(heap*rfile))(g' (H1,b) s)->(g (a,b) s)));
 [Clear H4; Intro H4 | Auto].
Inversion_clear H4.
Cut ((inv ((hwrite a (plus n (lookupR b r)) (lookupR b r0)),b))
  /\(p' ((hwrite a (plus n (lookupR b r)) (lookupR b r0)),b)))
  /\((s:(heap*rfile))(g' ((hwrite a (plus n (lookupR b r)) (lookupR b r0)),b) s)->(g (a,b) s)); [Clear H3; Intro H3; Inversion_clear H3 | Auto].
Apply wf2_w with t1 t2 inv A1 G1; Auto.
(** yield **)
Clear IHI2.
Inversion_clear H20; Inversion_clear H2.
Inversion_clear H3.
Rewrite <- H23 in H2.
Cut inv0==inv; [Intro; Clear H2 | Auto].
Rewrite H3 in H4; Rewrite H3 in H5; Clear H3 inv0.
Rewrite <- H23 in H8.
Cut (pairT A G)==(pairT A2 G2); [Clear H8; Intro | Auto].
Inversion H2; Clear H2.
Rewrite H8 in H5; Rewrite H9 in H6; Clear H8 H9 A G W.
Inversion_clear H7.
(*** case 1 ***)
Rewrite H2; Clear H2 i.
Exists p1; Exists p2; Exists G1; Apply wf1_w with t1 t2 inv A2 G2; Auto.
Intros; Apply H5 with s; Auto.
(*** case 2 ***)
Rewrite H2; Clear H2 i.
Exists p1; Exists p2; Exists G2; Apply wf2_w with t1 t2 inv A1 G1; Auto.
Intros; Apply H19 with s; Auto.
(** jd **)
Inversion_clear H20; Inversion_clear H2.
Inversion_clear H3.
Rewrite <- H23 in H7.
Cut inv0==inv; [Intro; Clear H7 | Auto].
Rewrite H3 in H5; Clear H3 inv0.
Rewrite <- H23 in H2.
Cut psi==psi2; [Clear H2; Intro | Auto].
Rewrite H2 in H4; Clear H2 psi0 W.
Exists p1; Exists p'; Exists g'.
Cut (p' s)/\((s':(heap*rfile))(g' s s')->(g s s'));
 [Clear H5; Intro H5; Inversion_clear H5 | Auto].
Apply wf2_w with t1 t2 inv A1 G1; Auto.
Rewrite <- H23.
Apply CT_CH_Infer with ch2 n; Auto.
Replace (getT2
       (pairT (pairT inv (pairT psi1 psi2))
         (pairT (pairT A1 G1) (pairT A2 G2)))) with t2; Auto.
Save.
