(****** CCAP Implementation for Dekker's Algorithm ******)
(****** created Mar. 26 2004, revised Jun. 29 2004 ******)

Require Arith.
Require ccap.

Syntactic Definition cs1 := (1).
Syntactic Definition cs2 := (2).
Syntactic Definition a1 := (3).
Syntactic Definition a2 := (4).

Definition H_dekker : heap :=
 [x:addr]
 (if (beq_nat x cs1) then (hvsome O) else
   (if (beq_nat x cs2) then (hvsome O) else
     (if (beq_nat x a1) then (hvsome O) else
       (if (beq_nat x a2) then (hvsome O) else hvnone)))).
Definition S_dekker : state := (H_dekker, emptyR).

Syntactic Definition loop1 := (1).
Syntactic Definition csec1 := (2).
Syntactic Definition reset1 := (3).
Syntactic Definition loop2 := (4).
Syntactic Definition csec2 := (5).
Syntactic Definition reset2 := (6).

Definition I_loop1 : iseq :=
 (yield
   (seq (movi r1 a1)
     (seq (movi r2 (1))
       (st r1 O r2
	 (yield
	   (seq (movi r1 cs1)
	     (seq (movi r2 (1))
	       (seq (movi r3 a2)
		 (ld r4 r3 O
		   (seq (sub r5 r2 r4)
		     (st r1 O r5
		       (yield
			 (seq (movi r1 cs1)
			   (ld r2 r1 O
			     (seq (movi r3 O)
			       (bgt r2 r3 csec1
				 (yield
				   (jd reset1)))))))))))))))))).
Definition I_csec1 : iseq :=
 (yield
   (seq (movi r1 cs1)
     (seq (movi r2 O)
       (st r1 O r2
	 (yield
	   (jd reset1)))))).
Definition I_reset1 : iseq :=
 (yield
   (seq (movi r1 a1)
     (seq (movi r2 O)
       (st r1 O r2
	 (yield
	   (jd loop1)))))).
Definition C1_dekker : codeheap :=
 (consC loop1 I_loop1
   (consC csec1 I_csec1
     (consC reset1 I_reset1 emptyC))).

Definition I_loop2 : iseq :=
 (yield
   (seq (movi r1 a2)
     (seq (movi r2 (1))
       (st r1 O r2
	 (yield
	   (seq (movi r1 cs2)
	     (seq (movi r2 (1))
	       (seq (movi r3 a1)
		 (ld r4 r3 O
		   (seq (sub r5 r2 r4)
		     (st r1 O r5
		       (yield
			 (seq (movi r1 cs2)
			   (ld r2 r1 O
			     (seq (movi r3 O)
			       (bgt r2 r3 csec2
				 (yield
				   (jd reset2)))))))))))))))))).
Definition I_csec2 : iseq :=
 (yield
   (seq (movi r1 cs2)
     (seq (movi r2 O)
       (st r1 O r2
	 (yield
	   (jd reset2)))))).
Definition I_reset2 : iseq :=
 (yield
   (seq (movi r1 a2)
     (seq (movi r2 O)
       (st r1 O r2
	 (yield
	   (jd loop2)))))).
Definition C2_dekker : codeheap :=
 (consC loop2 I_loop2
   (consC csec2 I_csec2
     (consC reset2 I_reset2 emptyC))).

Definition Prog_dekker : prog
  := ((S_dekker, (C1_dekker, C2_dekker)), (((jd loop1), (jd loop2)), (1))).

Definition Inv_dekker : assert :=
 [s:state]
  let (H, R) = s in (EX vcs1 | (EX vcs2 | (EX va1 | (EX va2 |
	 ((hread H cs1) = (hvsome vcs1))
	  /\ ((hread H cs2) = (hvsome vcs2))
	  /\ ((hread H a1) = (hvsome va1))
	  /\ ((hread H a2) = (hvsome va2))
	  /\ (le vcs1 (1)) /\ (le vcs2 (1)) /\ (le va1 (1)) /\ (le va2 (1))
	  /\ ~(vcs1 = (1) /\ vcs2 = (1))
	  /\ (vcs1 = (1) -> va1 = (1)) /\ (vcs2 = (1) -> va2 = (1)))))).

Definition A1_dekker : guar :=
 [s,s':state]
  let (H, R) = s in let (H', R') = s' in
 ((hread H a1) = (hread H' a1)) /\ ((hread H cs1) = (hread H' cs1)).
Definition A2_dekker : guar :=
 [s,s':state]
  let (H, R) = s in let (H', R') = s' in
 ((hread H a2) = (hread H' a2)) /\ ((hread H cs2) = (hread H' cs2)).
Definition G1_dekker : guar := A2_dekker.
Definition G2_dekker : guar := A1_dekker.

Definition a_loop1 : assert := [s:state] True.
Definition g_loop1 : guar := G1_dekker.
Definition a_csec1 : assert := [s:state] True.
Definition g_csec1 : guar := G1_dekker.
Definition a_reset1 : assert := [s:state] (getValinH s cs1)=(hvsome O).
Definition g_reset1 : guar := G1_dekker.
Definition CT1_dekker : codeheapty :=
 (consCT loop1 a_loop1 g_loop1
   (consCT csec1 a_csec1 g_csec1
     (consCT reset1 a_reset1 g_reset1 emptyCT))).

Definition a_loop2 : assert := [s:state] True.
Definition g_loop2 : guar := G2_dekker.
Definition a_csec2 : assert := [s:state] True.
Definition g_csec2 : guar := G2_dekker.
Definition a_reset2 : assert := [s:state] (getValinH s cs2)=(hvsome O).
Definition g_reset2 : guar := G2_dekker.
Definition CT2_dekker : codeheapty :=
 (consCT loop2 a_loop2 g_loop2
   (consCT csec2 a_csec2 g_csec2
     (consCT reset2 a_reset2 g_reset2 emptyCT))).

Definition Spec_dekker :=
 (pairT
   (pairT Inv_dekker (pairT CT1_dekker CT2_dekker))
   (pairT (pairT A1_dekker G1_dekker) (pairT A2_dekker G2_dekker))).

Definition TSpec1_dekker :=
 (pairT
   (pairT Inv_dekker CT1_dekker)
   (pairT A1_dekker G1_dekker)).
Definition TSpec2_dekker :=
 (pairT
   (pairT Inv_dekker CT2_dekker)
   (pairT A2_dekker G2_dekker)).

Definition p1_dekker : assert := [s:state] True.
Definition p2_dekker : assert := [s:state] True.
Definition g1_dekker : guar := G1_dekker.

Lemma sim_neg : (x:nat)
 (le x (1)) -> (le (minus (1) x) (1)).
Proof.
Intros.
Inversion_clear H; Auto.
Inversion_clear H0; Auto.
Save.

Lemma Infer_loop1 : (Infer TSpec1_dekker a_loop1 g_loop1 I_loop1).
Proof.
Unfold a_loop1; Unfold g_loop1; Unfold I_loop1.
(* yield *)
Apply I_yld with Inv_dekker A1_dekker G1_dekker; Auto.
Intros; Induction s; Unfold G1_dekker.
Unfold A2_dekker; Auto.
(* movi r1 a1 *)
Apply I_seq with Inv_dekker ([s:state] (getValinR s r1)=a1) G1_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* movi r2 1 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinR s r1)=a1 /\ (getValinR s r2)=(1))
  G1_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* st r1 (0) r2 *)
Apply I_st with Inv_dekker
 ([s:state] (getValinH s a1)=(hvsome (1)))
  G1_dekker; Auto.
Unfold getValinR; Unfold Inv_dekker; Intros.
SplitAllHyps.
SplitAllExs.
SplitAllHyps.
Unfold plus; Rewrite H0; Rewrite H3; Constructor.
Apply indom_some with x1; Auto.
Intros; Constructor.
Rewrite <- H14; Constructor; Auto.
Exists x; Exists x0; Exists (1); Exists x2.
Rewrite <- H2; Rewrite <- H1; Rewrite <- H5.
Repeat (Constructor; Auto).
Intro; Induction s; Unfold G1_dekker; Unfold A2_dekker; Intro.
Inversion_clear H15.
Rewrite <- H16.
Rewrite <- H17.
Rewrite <- H14.
Constructor; Auto.
(* yield *)
Apply I_yld with Inv_dekker A1_dekker G1_dekker; Auto.
Intros; Induction s; Unfold G1_dekker.
Unfold A2_dekker; Auto.
Intros; Induction s; Induction s'.
Unfold getValinH in H; Unfold getValinH.
Inversion_clear H.
Unfold A1_dekker in H0; Unfold Inv_dekker in H1.
SplitAllExs; SplitAllHyps.
Rewrite <- H11; Auto.
(* movi r1 cs1 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinH s a1)=(hvsome (1)) /\ (getValinR s r1)=cs1)
  G1_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* movi r2 1 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinH s a1) = (hvsome (1))
  /\ (getValinR s r1)=cs1 /\ (getValinR s r2)=(1))
  G1_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
SplitAllHyps.
Constructor; Auto.
(* movi r3 a2 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinH s a1)=(hvsome (1)) /\ (getValinR s r1)=cs1
  /\ (getValinR s r2)=(1) /\ (getValinR s r3)=a2)
  G1_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; SplitAllHyps.
Inversion_clear H.
Constructor; Auto.
(* ld r4 r3 (0) *)
Apply I_ld with Inv_dekker
 ([s:state] (getValinH s a1)=(hvsome (1)) /\(getValinR s r1)=cs1
  /\ (getValinR s r2)=(1) /\ (hvsome (getValinR s r4))=(getValinH s a2))
  G1_dekker; Auto.
Unfold getValinR; Unfold getValinH; Unfold Inv_dekker; Intros.
SplitAllHyps.
SplitAllExs.
SplitAllHyps.
Unfold plus.
Rewrite H5; Rewrite H4; Rewrite H1; Rewrite H6; Rewrite H7.
Exists x2; Constructor; Auto.
Rewrite H0 in H6; Constructor; Auto.
Exists x; Exists x0; Exists x1; Exists x2.
Repeat (Constructor; Auto).
(* sub r5 r2 r4 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinH s a1)=(hvsome (1)) /\ (getValinR s r1)=cs1
  /\ (getValinR s r5)=(minus (1) (getValinR s r4))
  /\ (hvsome (getValinR s r4))=(getValinH s a2))
  G1_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; SplitAllHyps.
Inversion_clear H.
Unfold getValinR; Unfold getValinR in H2; Unfold getValinR in H3.
Repeat (Constructor; Auto).
Rewrite H3; Auto.
(* st r1 (0) r5 *)
Apply I_st with Inv_dekker ([s:state] True) G1_dekker; Auto.
Unfold getValinR; Unfold getValinH; Unfold Inv_dekker; Intros.
SplitAllHyps.
SplitAllExs.
SplitAllHyps.
Unfold plus; Rewrite H2.
Constructor.
Apply indom_some with x; Auto.
Intros.
Rewrite <- H16; Constructor; Auto.
Constructor; Auto.
Exists (lookupR R r5); Exists x0; Exists x1; Exists x2.
Constructor; Auto.
Constructor; Auto.
Constructor; Auto.
Constructor; Auto.
Rewrite H3; Rewrite H7 in H5.
Inversion H5.
Rewrite H18; Constructor.
Apply sim_neg; Assumption.
Repeat (Constructor; Auto).
Intro HH; Inversion_clear HH.
Absurd x2=(1); Auto.
Intro.
Rewrite H20 in H17; Inversion H17.
Rewrite H0 in H6.
Inversion H6; Auto.
(* yield *)
Apply I_yld with Inv_dekker A1_dekker G1_dekker; Auto.
Intros; Induction s; Unfold G1_dekker.
Unfold A2_dekker; Auto.
(* movi r1 cs1 *)
Apply I_seq with Inv_dekker ([s:state] (getValinR s r1)=cs1) G1_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* ld r2 r1 (0) *)
Apply I_ld with Inv_dekker
 ([s:state] (hvsome (getValinR s r2))=(getValinH s cs1))
  G1_dekker; Auto.
Unfold getValinR; Unfold getValinH; Unfold Inv_dekker; Intros.
SplitAllHyps.
SplitAllExs.
SplitAllHyps.
Unfold plus.
Rewrite H0; Rewrite H1; Rewrite H3; Rewrite H4; Rewrite H2.
Exists x; Constructor; Auto.
Constructor; Auto.
Exists x; Exists x0; Exists x1; Exists x2.
Repeat (Constructor; Auto).
(* movi r3 0 *)
Apply I_seq with Inv_dekker
 ([s:state] (hvsome (getValinR s r2))=(getValinH s cs1) /\ (getValinR s r3)=O)
  G1_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* bgt r2 r3 csec1 *)
Apply I_bgt with CT1_dekker Inv_dekker
  a_csec1 ([s:state] (getValinH s cs1)=(hvsome O)) g_csec1 G1_dekker; Auto.
Unfold CT1_dekker.
LookupCT.
Unfold getValinR; Unfold a_csec1; Auto.
Unfold getValinR; Unfold getValinH; Intros.
SplitAllHyps.
Constructor; Auto.
Rewrite <- H1.
Replace (lookupR R r2) with O; Auto.
Rewrite H4 in H0.
Inversion_clear H0; Auto.
(* yield *)
Apply I_yld with Inv_dekker A1_dekker G1_dekker; Auto.
Intros; Induction s; Unfold G1_dekker.
Unfold A2_dekker; Auto.
Intros; Induction s; Induction s'.
Unfold getValinH in H; Unfold getValinH; Unfold A1_dekker in H0.
SplitAllHyps.
Rewrite <- H3; Auto.
(* jd reset1 *)
Apply I_jd with CT1_dekker Inv_dekker a_reset1 g_reset1; Auto.
Unfold CT1_dekker.
LookupCT.
Intros; Induction s.
Inversion_clear H; Unfold a_reset1; Constructor; Auto.
Save.

Lemma Infer_csec1 : (Infer TSpec1_dekker a_csec1 g_csec1 I_csec1).
Proof.
Unfold a_csec1; Unfold g_csec1; Unfold I_csec1.
(* yield *)
Apply I_yld with Inv_dekker A1_dekker G1_dekker; Auto.
Intros; Induction s; Unfold G1_dekker.
Unfold A2_dekker; Auto.
(* movi r1 cs1 *)
Apply I_seq with Inv_dekker ([s:state] (getValinR s r1)=cs1) G1_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* movi r2 0 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinR s r1)=cs1 /\ (getValinR s r2)=O)
  G1_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* st r1 (0) r2 *)
Apply I_st with Inv_dekker
 ([s:state] (getValinH s cs1)=(hvsome O))
  G1_dekker; Auto.
Unfold getValinR; Unfold Inv_dekker; Intros.
SplitAllHyps.
SplitAllExs.
SplitAllHyps.
Unfold plus; Rewrite H0; Rewrite H3; Constructor.
Apply indom_some with x; Auto.
Intros; Constructor.
Rewrite <- H14; Constructor; Auto.
Exists O; Exists x0; Exists x1; Exists x2.
Rewrite <- H1; Rewrite <- H4; Rewrite <- H5.
Repeat (Constructor; Auto).
Intro HH; Inversion_clear HH.
Inversion H15.
Intro HH; Inversion HH.
Intro; Induction s; Unfold G1_dekker; Unfold A2_dekker; Intro.
Rewrite <- H14 in H15.
Inversion_clear H15.
Rewrite <- H16.
Rewrite <- H17.
Constructor; Auto.
(* yield *)
Apply I_yld with Inv_dekker A1_dekker G1_dekker; Auto.
Intros; Induction s; Unfold G1_dekker.
Unfold A2_dekker; Auto.
Intros; Induction s; Induction s'.
Unfold getValinH in H; Unfold getValinH; Unfold A1_dekker in H0.
SplitAllHyps.
Rewrite <- H3; Auto.
(* jd reset1 *)
Apply I_jd with CT1_dekker Inv_dekker a_reset1 g_reset1; Auto.
Unfold CT1_dekker.
LookupCT.
Intros; Induction s.
Inversion_clear H; Unfold a_reset1; Constructor; Auto.
Save.

Lemma Infer_reset1 : (Infer TSpec1_dekker a_reset1 g_reset1 I_reset1).
Proof.
Unfold a_reset1; Unfold g_reset1; Unfold I_reset1.
(* yield *)
Apply I_yld with Inv_dekker A1_dekker G1_dekker; Auto.
Intros; Induction s; Unfold G1_dekker.
Unfold A2_dekker; Auto.
Intros; Induction s; Induction s'.
Unfold getValinH in H; Unfold getValinH; Unfold A1_dekker in H0.
SplitAllHyps.
Rewrite <- H3; Auto.
(* movi r1 a1 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinH s cs1)=(hvsome O) /\ (getValinR s r1)=a1)
  G1_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* movi r2 0 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinH s cs1)=(hvsome O)
  /\ (getValinR s r1)=a1 /\ (getValinR s r2)=O)
  G1_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
SplitAllHyps.
Constructor; Auto.
(* st r1 (0) r2 *)
Apply I_st with Inv_dekker ([s:state] True) G1_dekker; Auto.
Unfold getValinR; Unfold Inv_dekker; Intros.
SplitAllHyps.
SplitAllExs.
SplitAllHyps.
Unfold getValinH in H0; Unfold plus; Rewrite H2; Rewrite H4; Constructor.
Apply indom_some with x1; Auto.
Intros; Constructor.
Rewrite <- H15; Constructor; Auto.
Exists x; Exists x0; Exists O; Exists x2.
Rewrite <- H3; Rewrite <- H1; Rewrite <- H6.
Repeat (Constructor; Auto).
Rewrite H0 in H3; Inversion_clear H3; Auto.
Intro; Induction s; Unfold G1_dekker; Unfold A2_dekker; Intro.
Rewrite <- H15 in H16.
Inversion_clear H16.
Rewrite <- H17.
Rewrite <- H18.
Constructor; Auto.
(* yield *)
Apply I_yld with Inv_dekker A1_dekker G1_dekker; Auto.
Intros; Induction s; Unfold G1_dekker.
Unfold A2_dekker; Auto.
(* jd loop1 *)
Apply I_jd with CT1_dekker Inv_dekker a_loop1 g_loop1; Auto.
Unfold CT1_dekker.
LookupCT.
Intros; Induction s.
Unfold a_loop1; Constructor; Auto.
Save.

Lemma WF_CH1_dekker : (WF_CH TSpec1_dekker C1_dekker).
Proof.
Apply wf_c.
Replace (tgetCT TSpec1_dekker) with CT1_dekker;
 [Unfold CT1_dekker; Unfold C1_dekker | Auto].
Apply wf_c_cons.
Apply Infer_loop1.
Apply wf_c_cons.
Apply Infer_csec1.
Apply wf_c_cons.
Apply Infer_reset1.
Apply wf_c_emp.
Save.

Lemma Infer_loop2 : (Infer TSpec2_dekker a_loop2 g_loop2 I_loop2).
Proof.
Unfold a_loop2; Unfold g_loop2; Unfold I_loop2.
(* yield *)
Apply I_yld with Inv_dekker A2_dekker G2_dekker; Auto.
Intros; Induction s; Unfold G2_dekker.
Unfold A1_dekker; Auto.
(* movi r1 a2 *)
Apply I_seq with Inv_dekker ([s:state] (getValinR s r1)=a2) G2_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* movi r2 1 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinR s r1)=a2 /\ (getValinR s r2)=(1))
  G2_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* st r1 (0) r2 *)
Apply I_st with Inv_dekker
 ([s:state] (getValinH s a2)=(hvsome (1)))
  G2_dekker; Auto.
Unfold getValinR; Unfold Inv_dekker; Intros.
SplitAllHyps.
SplitAllExs.
SplitAllHyps.
Unfold plus; Rewrite H0; Rewrite H3; Constructor.
Apply indom_some with x2; Auto.
Intros; Constructor.
Rewrite <- H14; Constructor; Auto.
Exists x; Exists x0; Exists x1; Exists (1).
Rewrite <- H2; Rewrite <- H1; Rewrite <- H4.
Repeat (Constructor; Auto).
Intro; Induction s; Unfold G2_dekker; Unfold A1_dekker; Intro.
Inversion_clear H15.
Rewrite <- H16.
Rewrite <- H17.
Rewrite <- H14.
Constructor; Auto.
(* yield *)
Apply I_yld with Inv_dekker A2_dekker G2_dekker; Auto.
Intros; Induction s; Unfold G2_dekker.
Unfold A1_dekker; Auto.
Intros; Induction s; Induction s'.
Unfold getValinH in H; Unfold getValinH.
Inversion_clear H.
Unfold A2_dekker in H0; Unfold Inv_dekker in H1.
SplitAllExs; SplitAllHyps.
Rewrite <- H11; Auto.
(* movi r1 cs2 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinH s a2)=(hvsome (1)) /\ (getValinR s r1)=cs2)
  G2_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* movi r2 1 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinH s a2) = (hvsome (1))
  /\ (getValinR s r1)=cs2 /\ (getValinR s r2)=(1))
  G2_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
SplitAllHyps.
Constructor; Auto.
(* movi r3 a1 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinH s a2)=(hvsome (1)) /\ (getValinR s r1)=cs2
  /\ (getValinR s r2)=(1) /\ (getValinR s r3)=a1)
  G2_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; SplitAllHyps.
Inversion_clear H.
Constructor; Auto.
(* ld r4 r3 (0) *)
Apply I_ld with Inv_dekker
 ([s:state] (getValinH s a2)=(hvsome (1)) /\(getValinR s r1)=cs2
  /\ (getValinR s r2)=(1) /\ (hvsome (getValinR s r4))=(getValinH s a1))
  G2_dekker; Auto.
Unfold getValinR; Unfold getValinH; Unfold Inv_dekker; Intros.
SplitAllHyps.
SplitAllExs.
SplitAllHyps.
Unfold plus.
Rewrite H5; Rewrite H4; Rewrite H1; Rewrite H6; Rewrite H7.
Exists x1; Constructor; Auto.
Rewrite H0 in H7; Constructor; Auto.
Exists x; Exists x0; Exists x1; Exists x2.
Repeat (Constructor; Auto).
(* sub r5 r2 r4 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinH s a2)=(hvsome (1)) /\ (getValinR s r1)=cs2
  /\ (getValinR s r5)=(minus (1) (getValinR s r4))
  /\ (hvsome (getValinR s r4))=(getValinH s a1))
  G2_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; SplitAllHyps.
Inversion_clear H.
Unfold getValinR; Unfold getValinR in H2; Unfold getValinR in H3.
Repeat (Constructor; Auto).
Rewrite H3; Auto.
(* st r1 (0) r5 *)
Apply I_st with Inv_dekker ([s:state] True) G2_dekker; Auto.
Unfold getValinR; Unfold getValinH; Unfold Inv_dekker; Intros.
SplitAllHyps.
SplitAllExs.
SplitAllHyps.
Unfold plus; Rewrite H2.
Constructor.
Apply indom_some with x0; Auto.
Intros.
Rewrite <- H16; Constructor; Auto.
Constructor; Auto.
Exists x; Exists (lookupR R r5); Exists x1; Exists x2.
Constructor; Auto.
Constructor; Auto.
Constructor; Auto.
Constructor; Auto.
Rewrite H3; Rewrite H6 in H5.
Inversion H5.
Rewrite H18; Constructor; Auto.
Constructor.
Apply sim_neg; Assumption.
Repeat (Constructor; Auto).
Intro HH; Inversion_clear HH.
Absurd x1=(1); Auto.
Intro.
Rewrite H20 in H19; Inversion H19.
Rewrite H0 in H7.
Inversion H7; Auto.
(* yield *)
Apply I_yld with Inv_dekker A2_dekker G2_dekker; Auto.
Intros; Induction s; Unfold G2_dekker.
Unfold A1_dekker; Auto.
(* movi r1 cs2 *)
Apply I_seq with Inv_dekker ([s:state] (getValinR s r1)=cs2) G2_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* ld r2 r1 (0) *)
Apply I_ld with Inv_dekker
 ([s:state] (hvsome (getValinR s r2))=(getValinH s cs2))
  G2_dekker; Auto.
Unfold getValinR; Unfold getValinH; Unfold Inv_dekker; Intros.
SplitAllHyps.
SplitAllExs.
SplitAllHyps.
Unfold plus.
Rewrite H0; Rewrite H1; Rewrite H3; Rewrite H4; Rewrite H2.
Exists x0; Constructor; Auto.
Constructor; Auto.
Exists x; Exists x0; Exists x1; Exists x2.
Repeat (Constructor; Auto).
(* movi r3 0 *)
Apply I_seq with Inv_dekker
 ([s:state] (hvsome (getValinR s r2))=(getValinH s cs2) /\ (getValinR s r3)=O)
  G2_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* bgt r2 r3 csec2 *)
Apply I_bgt with CT2_dekker Inv_dekker
  a_csec2 ([s:state] (getValinH s cs2)=(hvsome O)) g_csec2 G2_dekker; Auto.
Unfold CT2_dekker.
LookupCT.
Unfold getValinR; Unfold a_csec2; Auto.
Unfold getValinR; Unfold getValinH; Intros.
SplitAllHyps.
Constructor; Auto.
Rewrite <- H1.
Replace (lookupR R r2) with O; Auto.
Rewrite H4 in H0.
Inversion_clear H0; Auto.
(* yield *)
Apply I_yld with Inv_dekker A2_dekker G2_dekker; Auto.
Intros; Induction s; Unfold G2_dekker.
Unfold A1_dekker; Auto.
Intros; Induction s; Induction s'.
Unfold getValinH in H; Unfold getValinH; Unfold A2_dekker in H0.
SplitAllHyps.
Rewrite <- H3; Auto.
(* jd reset2 *)
Apply I_jd with CT2_dekker Inv_dekker a_reset2 g_reset2; Auto.
Unfold CT2_dekker.
LookupCT.
Intros; Induction s.
Inversion_clear H; Unfold a_reset2; Constructor; Auto.
Save.

Lemma Infer_csec2 : (Infer TSpec2_dekker a_csec2 g_csec2 I_csec2).
Proof.
Unfold a_csec2; Unfold g_csec2; Unfold I_csec2.
(* yield *)
Apply I_yld with Inv_dekker A2_dekker G2_dekker; Auto.
Intros; Induction s; Unfold G2_dekker.
Unfold A1_dekker; Auto.
(* movi r1 cs2 *)
Apply I_seq with Inv_dekker ([s:state] (getValinR s r1)=cs2) G2_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* movi r2 0 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinR s r1)=cs2 /\ (getValinR s r2)=O)
  G2_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* st r1 (0) r2 *)
Apply I_st with Inv_dekker
 ([s:state] (getValinH s cs2)=(hvsome O))
  G2_dekker; Auto.
Unfold getValinR; Unfold Inv_dekker; Intros.
SplitAllHyps.
SplitAllExs.
SplitAllHyps.
Unfold plus; Rewrite H0; Rewrite H3; Constructor.
Apply indom_some with x0; Auto.
Intros; Constructor.
Rewrite <- H14; Constructor; Auto.
Exists x; Exists O; Exists x1; Exists x2.
Rewrite <- H2; Rewrite <- H4; Rewrite <- H5.
Repeat (Constructor; Auto).
Intro HH; Inversion_clear HH.
Inversion H16.
Intro HH; Inversion HH.
Intro; Induction s; Unfold G2_dekker; Unfold A1_dekker; Intro.
Rewrite <- H14 in H15.
Inversion_clear H15.
Rewrite <- H16.
Rewrite <- H17.
Constructor; Auto.
(* yield *)
Apply I_yld with Inv_dekker A2_dekker G2_dekker; Auto.
Intros; Induction s; Unfold G2_dekker.
Unfold A1_dekker; Auto.
Intros; Induction s; Induction s'.
Unfold getValinH in H; Unfold getValinH; Unfold A2_dekker in H0.
SplitAllHyps.
Rewrite <- H3; Auto.
(* jd reset2 *)
Apply I_jd with CT2_dekker Inv_dekker a_reset2 g_reset2; Auto.
Unfold CT2_dekker.
LookupCT.
Intros; Induction s.
Inversion_clear H; Unfold a_reset2; Constructor; Auto.
Save.

Lemma Infer_reset2 : (Infer TSpec2_dekker a_reset2 g_reset2 I_reset2).
Proof.
Unfold a_reset2; Unfold g_reset2; Unfold I_reset2.
(* yield *)
Apply I_yld with Inv_dekker A2_dekker G2_dekker; Auto.
Intros; Induction s; Unfold G2_dekker.
Unfold A1_dekker; Auto.
Intros; Induction s; Induction s'.
Unfold getValinH in H; Unfold getValinH; Unfold A2_dekker in H0.
SplitAllHyps.
Rewrite <- H3; Auto.
(* movi r1 a2 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinH s cs2)=(hvsome O) /\ (getValinR s r1)=a2)
  G2_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
Inversion_clear H0; Constructor; Auto.
(* movi r2 0 *)
Apply I_seq with Inv_dekker
 ([s:state] (getValinH s cs2)=(hvsome O)
  /\ (getValinR s r1)=a2 /\ (getValinR s r2)=O)
  G2_dekker; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H.
Inversion_clear H.
SplitAllHyps.
Constructor; Auto.
(* st r1 (0) r2 *)
Apply I_st with Inv_dekker ([s:state] True) G2_dekker; Auto.
Unfold getValinR; Unfold Inv_dekker; Intros.
SplitAllHyps.
SplitAllExs.
SplitAllHyps.
Unfold getValinH in H0; Unfold plus; Rewrite H2; Rewrite H4; Constructor.
Apply indom_some with x2; Auto.
Intros; Constructor.
Rewrite <- H15; Constructor; Auto.
Exists x; Exists x0; Exists x1; Exists O.
Rewrite <- H3; Rewrite <- H1; Rewrite <- H5.
Repeat (Constructor; Auto).
Rewrite H0 in H1; Inversion_clear H1; Auto.
Intro; Induction s; Unfold G2_dekker; Unfold A1_dekker; Intro.
Rewrite <- H15 in H16.
Inversion_clear H16.
Rewrite <- H17.
Rewrite <- H18.
Constructor; Auto.
(* yield *)
Apply I_yld with Inv_dekker A2_dekker G2_dekker; Auto.
Intros; Induction s; Unfold G2_dekker.
Unfold A1_dekker; Auto.
(* jd loop2 *)
Apply I_jd with CT2_dekker Inv_dekker a_loop2 g_loop2; Auto.
Unfold CT2_dekker.
LookupCT.
Intros; Induction s.
Unfold a_loop2; Constructor; Auto.
Save.

Lemma WF_CH2_dekker : (WF_CH TSpec2_dekker C2_dekker).
Proof.
Apply wf_c.
Replace (tgetCT TSpec2_dekker) with CT2_dekker;
 [Unfold CT2_dekker; Unfold C2_dekker | Auto].
Apply wf_c_cons.
Apply Infer_loop2.
Apply wf_c_cons.
Apply Infer_csec2.
Apply wf_c_cons.
Apply Infer_reset2.
Apply wf_c_emp.
Save.

Theorem WF_dekker : (WFW Spec_dekker p1_dekker p2_dekker g1_dekker Prog_dekker).
Proof.
Unfold Prog_dekker.
Apply wf1_w with TSpec1_dekker TSpec2_dekker Inv_dekker A2_dekker G2_dekker;
  Auto.
Apply WF_CH1_dekker.
Apply WF_CH2_dekker.
Unfold Inv_dekker; Unfold S_dekker; Unfold p1_dekker; Constructor; Auto.
Unfold H_dekker; Exists O; Exists O; Exists O; Exists O.
Repeat (Constructor; Auto).
Intro FF; Inversion_clear FF.
Inversion H.
Intro; Unfold p2_dekker; Auto.
Intros; Unfold p2_dekker; Auto.
Apply I_jd with CT1_dekker Inv_dekker ([s:state] True) G1_dekker; Auto.
Unfold CT1_dekker.
LookupCT.
Apply I_jd with CT2_dekker Inv_dekker ([s:state] True) G2_dekker; Auto.
Unfold CT2_dekker.
LookupCT.
Save.
