(****** CCAP Implementation for Greated Common Divisor ******)
(******     created Mar. 29 2004, revised Jun. 29 2004 ******)

Require Arith.
Require ccap.

(* Rigid variables alpha and beta. *)
Parameter alpha : nat.
Parameter beta : nat.

(* Any correct implementation of gcd should satisfy the properties. *)
Parameter gcd : nat -> nat -> nat.
Axiom gcd_n_n : (n:nat) n=(gcd n n).
Axiom gcd_minus : (n,m:nat)
 (gt n m) -> (gcd (minus n m) m)=(gcd n m).
Axiom gcd_sym : (n,m:nat) (gcd n m)=(gcd m n).

Syntactic Definition a := (1).
Syntactic Definition b := (2).

Definition H_gcd : heap :=
 [x:addr]
 (if (beq_nat x a) then (hvsome alpha) else
   (if (beq_nat x b) then (hvsome beta) else hvnone)).
Definition S_gcd : state := (H_gcd, emptyR).

Syntactic Definition loop1 := (1).
Syntactic Definition calc1 := (2).
Syntactic Definition done1 := (3).
Syntactic Definition loop2 := (4).
Syntactic Definition calc2 := (5).
Syntactic Definition done2 := (6).

Definition I_loop1 : iseq :=
 (yield
   (seq (movi r1 a)
     (ld r2 r1 O
       (seq (movi r3 b)
	 (ld r4 r3 O
	   (be r2 r4 done1
	     (yield
	       (seq (movi r1 a)
		 (ld r2 r1 O
		   (seq (movi r3 b)
		     (ld r4 r3 O
		       (bgt r2 r4 calc1
			 (yield
			   (jd loop1)))))))))))))).
Definition I_calc1 : iseq :=
 (yield
   (seq (movi r1 a)
     (ld r2 r1 O
       (seq (movi r3 b)
	 (ld r4 r3 O
	   (seq (sub r5 r2 r4)
	     (st r1 O r5
	       (yield
		 (jd loop1))))))))).
Definition I_done1 : iseq :=
 (yield
   (jd done1)).
Definition C1_gcd : codeheap :=
 (consC loop1 I_loop1
   (consC calc1 I_calc1
     (consC done1 I_done1 emptyC))).

Definition I_loop2 : iseq :=
 (yield
   (seq (movi r1 b)
     (ld r2 r1 O
       (seq (movi r3 a)
	 (ld r4 r3 O
	   (be r2 r4 done2
	     (yield
	       (seq (movi r1 b)
		 (ld r2 r1 O
		   (seq (movi r3 a)
		     (ld r4 r3 O
		       (bgt r2 r4 calc2
			 (yield
			   (jd loop2)))))))))))))).
Definition I_calc2 : iseq :=
 (yield
   (seq (movi r1 b)
     (ld r2 r1 O
       (seq (movi r3 a)
	 (ld r4 r3 O
	   (seq (sub r5 r2 r4)
	     (st r1 O r5
	       (yield
		 (jd loop2))))))))).
Definition I_done2 : iseq :=
 (yield
   (jd done2)).
Definition C2_gcd : codeheap :=
 (consC loop2 I_loop2
   (consC calc2 I_calc2
     (consC done2 I_done2 emptyC))).

Definition Prog_gcd : prog :=
 ((S_gcd, (C1_gcd, C2_gcd)), (((jd loop1), (jd loop2)), (1))).

Definition Inv_gcd : assert :=
 [s:state] (EX va | (EX vb |
     (getValinH s a) = (hvsome va) /\ (getValinH s b) = (hvsome vb)
      /\ (gcd va vb) = (gcd alpha beta))).

Definition A1_gcd : guar :=
 [s,s':state] (EX va | (EX vb |
     (getValinH s a)=(hvsome va)
      /\ (getValinH s b)=(hvsome vb)
      /\ ((le vb va) ->
          ((getValinH s' a)=(hvsome va) /\ (getValinH s' b)=(hvsome vb)))
      /\ ((gt vb va) ->
	  (EX va' | (EX vb' |
	      (getValinH s' a)=(hvsome va')
	       /\ (getValinH s' b)=(hvsome vb')
	       /\ ((gcd va' vb')=(gcd alpha beta))))))).
Definition A2_gcd : guar :=
 [s,s':state] (EX vb | (EX va |
     (getValinH s b)=(hvsome vb)
      /\ (getValinH s a)=(hvsome va)
      /\ ((le va vb) ->
          ((getValinH s' b)=(hvsome vb) /\ (getValinH s' a)=(hvsome va)))
      /\ ((gt va vb) ->
	  (EX vb' | (EX va' |
	      (getValinH s' b)=(hvsome vb')
	       /\ (getValinH s' a)=(hvsome va')
	       /\ ((gcd vb' va')=(gcd alpha beta))))))).
Definition G1_gcd : guar := A2_gcd.
Definition G2_gcd : guar := A1_gcd.

Definition a_loop1 : assert := [s:state] True.
Definition g_loop1 : guar := G1_gcd.
Definition a_calc1 : assert :=
 [s:state] (EX va | (EX vb |
     (getValinH s a)=(hvsome va) /\ (getValinH s b)=(hvsome vb)
      /\ (gt va vb))).
Definition g_calc1 : guar := G1_gcd.
Definition a_done1 : assert :=
 [s:state] (EX v |
   (getValinH s a)=(hvsome v) /\ (getValinH s b)=(hvsome v)
    /\ v=(gcd alpha beta)).
Definition g_done1 : guar := G1_gcd.
Definition CT1_gcd : codeheapty :=
 (consCT loop1 a_loop1 g_loop1
   (consCT calc1 a_calc1 g_calc1
     (consCT done1 a_done1 g_done1 emptyCT))).

Definition a_loop2 : assert := [s:state] True.
Definition g_loop2 : guar := G2_gcd.
Definition a_calc2 : assert :=
 [s:state] (EX vb | (EX va |
     (getValinH s b)=(hvsome vb) /\ (getValinH s a)=(hvsome va)
      /\ (gt vb va))).
Definition g_calc2 : guar := G2_gcd.
Definition a_done2 : assert :=
 [s:state] (EX v |
   (getValinH s b)=(hvsome v) /\ (getValinH s a)=(hvsome v)
    /\ v=(gcd alpha beta)).
Definition g_done2 : guar := G2_gcd.
Definition CT2_gcd : codeheapty :=
 (consCT loop2 a_loop2 g_loop2
   (consCT calc2 a_calc2 g_calc2
     (consCT done2 a_done2 g_done2 emptyCT))).

Definition Spec_gcd :=
 (pairT
   (pairT Inv_gcd (pairT CT1_gcd CT2_gcd))
   (pairT (pairT A1_gcd G1_gcd) (pairT A2_gcd G2_gcd))).

Definition TSpec1_gcd :=
 (pairT
   (pairT Inv_gcd CT1_gcd)
   (pairT A1_gcd G1_gcd)).
Definition TSpec2_gcd :=
 (pairT
   (pairT Inv_gcd CT2_gcd)
   (pairT A2_gcd G2_gcd)).

Syntactic Definition p1_gcd := a_loop1.
Syntactic Definition p2_gcd := a_loop2.
Syntactic Definition g1_gcd := G1_gcd.

Lemma Infer_loop1 : (Infer TSpec1_gcd a_loop1 g_loop1 I_loop1).
Proof.
Unfold a_loop1; Unfold g_loop1; Unfold I_loop1.
(* yield *)
Apply I_yld with Inv_gcd A1_gcd G1_gcd; Auto.
Unfold G1_gcd; Unfold A2_gcd; Unfold Inv_gcd; Unfold getValinH; Intros; Induction s.
Inversion_clear H; SplitAllExs; SplitAllHyps.
Exists x0; Exists x; Repeat (Constructor; Auto).
Intro; Exists x0; Exists x; Repeat (Constructor; Auto).
Rewrite <- H3; Apply gcd_sym.
(* movi r1 a *)
Apply I_seq with Inv_gcd ([s:state] (getValinR s r1)=a) G1_gcd; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Inversion_clear H0; Auto.
(* ld r2 r1 (0) *)
Apply I_ld with Inv_gcd
 ([s:state] (hvsome (getValinR s r2)) = (getValinH s a))
  G1_gcd; Auto.
Unfold Inv_gcd; Unfold getValinH; Unfold getValinR; Unfold plus.
Intros; SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H2; Clear H2; Exists x; Repeat (Constructor; Auto).
Exists x; Exists x0; Auto.
(* movi r3 b *)
Apply I_seq with Inv_gcd
 ([s:state] (hvsome (getValinR s r2)) = (getValinH s a)
  /\ (getValinR s r3)=b)
  G1_gcd; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Inversion_clear H0; Auto.
(* ld r4 r3 (0) *)
Apply I_ld with Inv_gcd
 ([s:state] (hvsome (getValinR s r2)) = (getValinH s a)
  /\ (hvsome (getValinR s r4)) = (getValinH s b))
  G1_gcd; Auto.
Unfold Inv_gcd; Unfold getValinH; Unfold getValinR; Unfold plus.
Intros; SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H3; Clear H3; Exists x0; Repeat (Constructor; Auto).
Exists x; Exists x0; Auto.
(* be r2 r4 done1 *)
Apply I_be with CT1_gcd Inv_gcd a_done1 ([s:state] True) g_done1 G1_gcd; Auto.
Unfold CT1_gcd; LookupCT.
Unfold Inv_gcd; Unfold a_done1; Unfold g_done1; Unfold getValinH; Unfold getValinR; Intros.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Constructor; Auto.
Rewrite H0 in H1; Rewrite H1 in H4; Rewrite H3 in H4; Exists x.
Repeat (Constructor; Auto).
Rewrite H2 in H4; Inversion H4.
Rewrite H7 in H6; Rewrite <- H6; Apply gcd_n_n.
(* yield *)
Apply I_yld with Inv_gcd A1_gcd G1_gcd; Auto.
Unfold Inv_gcd; Unfold G1_gcd; Unfold A2_gcd; Unfold getValinH; Intros; Induction s.
Inversion_clear H; SplitAllExs; SplitAllHyps.
Exists x0; Exists x; Repeat (Constructor; Auto).
Intro; Exists x0; Exists x; Repeat (Constructor; Auto).
Rewrite <- H3; Apply gcd_sym.
(* movi r1 a *)
Apply I_seq with Inv_gcd ([s:state] (getValinR s r1)=a) G1_gcd; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Inversion_clear H0; Auto.
(* ld r2 r1 (0) *)
Apply I_ld with Inv_gcd
 ([s:state] (hvsome (getValinR s r2)) = (getValinH s a))
  G1_gcd; Auto.
Unfold Inv_gcd; Unfold getValinH; Unfold getValinR; Unfold plus.
Intros; SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H2; Clear H2; Exists x; Repeat (Constructor; Auto).
Exists x; Exists x0; Auto.
(* movi r3 b *)
Apply I_seq with Inv_gcd
 ([s:state] (hvsome (getValinR s r2)) = (getValinH s a)
  /\ (getValinR s r3)=b)
  G1_gcd; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Inversion_clear H0; Auto.
(* ld r4 r3 (0) *)
Apply I_ld with Inv_gcd
 ([s:state] (hvsome (getValinR s r2)) = (getValinH s a)
  /\ (hvsome (getValinR s r4)) = (getValinH s b))
  G1_gcd; Auto.
Unfold Inv_gcd; Unfold getValinH; Unfold getValinR; Unfold plus.
Intros; SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H3; Clear H3; Exists x0; Repeat (Constructor; Auto).
Exists x; Exists x0; Auto.
(* bgt r2 r4 calc1 *)
Apply I_bgt with CT1_gcd Inv_gcd a_calc1 ([s:state]True) g_calc1 G1_gcd; Auto.
Unfold CT1_gcd; LookupCT.
Unfold Inv_gcd; Unfold a_calc1; Unfold g_calc1; Unfold getValinH; Unfold getValinR; Intros.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Constructor; Auto.
Exists (lookupR R r2); Exists (lookupR R r4); Auto.
(* yield *)
Apply I_yld with Inv_gcd A1_gcd G1_gcd; Auto.
Unfold Inv_gcd; Unfold G1_gcd; Unfold A2_gcd; Unfold getValinH; Intros; Induction s.
Inversion_clear H; SplitAllExs; SplitAllHyps.
Exists x0; Exists x; Repeat (Constructor; Auto).
Intro; Exists x0; Exists x; Repeat (Constructor; Auto).
Rewrite <- H3; Apply gcd_sym.
(* jd loop1 *)
Apply I_jd with CT1_gcd Inv_gcd a_loop1 g_loop1; Auto.
Unfold CT1_gcd; LookupCT.
Intros; Induction s.
Inversion_clear H; Unfold a_loop1; Constructor; Auto.
Save.

Lemma Infer_calc1 : (Infer TSpec1_gcd a_calc1 g_calc1 I_calc1).
Proof.
Unfold a_calc1; Unfold g_calc1; Unfold I_calc1.
(* yield *)
Apply I_yld with Inv_gcd A1_gcd G1_gcd; Auto.
Unfold G1_gcd; Unfold A2_gcd; Unfold Inv_gcd; Unfold getValinH; Intros; Induction s.
Inversion_clear H; SplitAllExs; SplitAllHyps.
Exists x0; Exists x; Repeat (Constructor; Auto).
Intro; Exists x2; Exists x1; Repeat (Constructor; Auto).
Rewrite <- H3; Apply gcd_sym.
Unfold Inv_gcd; Unfold A1_gcd; Unfold getValinH; Intros; Induction s; Induction s'.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Clear H5 H4 H1 H7; Rewrite H in H6; Rewrite H0 in H2.
Exists x; Exists x0; Inversion H6; Inversion H2.
Rewrite H4 in H3; Rewrite H5 in H3.
Cut (hread a0 (1))=(Some nat x)/\(hread a0 (2))=(Some nat x0); Auto with arith.
Intro HH; Inversion_clear HH; Auto.
(* movi r1 a *)
Apply I_seq with Inv_gcd
 ([s:state] (EX va:nat | (EX vb:nat |
     (getValinH s a)=(hvsome va)
      /\ (getValinH s b)=(hvsome vb)
      /\ (gt va vb) /\ (getValinR s r1)=a)))
 ([s,s':state] (EX va':nat | (EX vb':nat |
     (getValinH s' a)=(hvsome va')
      /\ (getValinH s' b)=(hvsome vb')
      /\ (gcd va' vb')=(gcd alpha beta)))); Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Constructor; Auto.
Unfold getValinH in H0; Unfold getValinH in H1; Unfold getValinH; Constructor.
Exists x; Exists x0; Auto.
Intros; Induction s'; SplitAllExs; Unfold G1_gcd; Unfold A2_gcd.
Unfold getValinH; SplitAllHyps.
Exists x0; Exists x; Constructor; Auto.
Constructor; Auto.
Constructor; Intro.
Absurd (gt x x0); Auto with arith.
Exists x2; Exists x1; Repeat (Constructor; Auto).
Rewrite <- H6; Apply gcd_sym.
(* ld r2 r1 (0) *)
Apply I_ld with Inv_gcd
 ([s:state] (EX va:nat | (EX vb:nat |
       (getValinH s a)=(hvsome va)
	/\ (getValinH s b)=(hvsome vb)
	/\ (gt va vb) /\ (getValinR s r1)=a
	/\ (hvsome (getValinR s r2))=(getValinH s a))))
  ([s,s':state] (EX va':nat | (EX vb':nat |
     (getValinH s' a)=(hvsome va')
      /\ (getValinH s' b)=(hvsome vb')
      /\ (gcd va' vb')=(gcd alpha beta)))); Auto.
Unfold Inv_gcd; Unfold getValinH; Unfold getValinR; Unfold plus.
Intros; SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H7; Exists x; Repeat (Constructor; Auto).
Exists x1; Exists x2; Auto.
Exists x; Exists x0; Auto.
(* movi r3 b *)
Apply I_seq with Inv_gcd
 ([s:state] (EX va:nat | (EX vb:nat |
       (getValinH s a)=(hvsome va)
	/\ (getValinH s b)=(hvsome vb)
	/\ (gt va vb) /\ (getValinR s r1)=(1)
	/\ (hvsome (getValinR s r2))=(getValinH s a)
	/\(getValinR s r3)=b)))
 ([s,s':state] (EX va':nat | (EX vb':nat |
     (getValinH s' a)=(hvsome va')
      /\ (getValinH s' b)=(hvsome vb')
      /\ (gcd va' vb')=(gcd alpha beta)))); Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Repeat (Constructor; Auto).
Unfold getValinH in H0; Unfold getValinH in H1; Unfold getValinH.
Unfold getValinR in H3; Unfold getValinR.
Exists x; Exists x0; Repeat (Constructor; Auto).
(* ld r4 r3 (0) *)
Apply I_ld with Inv_gcd
 ([s:state] (EX va:nat | (EX vb:nat |
       (getValinH s a)=(hvsome va)
	/\ (getValinH s b)=(hvsome vb)
	/\ (gt va vb) /\ (getValinR s r1)=(1)
	/\ (hvsome (getValinR s r2))=(getValinH s a)
	/\(hvsome (getValinR s r4))=(getValinH s b))))
 ([s,s':state] (EX va':nat | (EX vb':nat |
     (getValinH s' a)=(hvsome va')
      /\ (getValinH s' b)=(hvsome vb')
      /\ (gcd va' vb')=(gcd alpha beta)))); Auto.
Unfold Inv_gcd; Unfold getValinH; Unfold getValinR; Unfold plus.
Intros; SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H9; Exists x0; Repeat (Constructor; Auto).
Exists x1; Exists x2; Auto.
Exists x; Exists x0; Repeat (Constructor; Auto).
(* sub r5 r2 r4 *)
Apply I_seq with Inv_gcd
 ([s:state] (EX va:nat | (EX vb:nat |
       (getValinH s a)=(hvsome va)
	/\ (getValinH s b)=(hvsome vb)
	/\ (gt va vb) /\ (getValinR s r1)=(1)
	/\ (getValinR s r5) = (minus va vb))))
 ([s,s':state] (EX va':nat | (EX vb':nat |
     (getValinH s' a)=(hvsome va')
      /\ (getValinH s' b)=(hvsome vb')
      /\ (gcd va' vb')=(gcd alpha beta)))); Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_gcd in H0; Unfold Inv_gcd; Unfold getValinH in H0; Unfold getValinR in H0; Unfold getValinH; Unfold getValinR.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Repeat (Constructor; Auto).
Exists x1; Exists x2; Auto.
Exists x; Exists x0; Repeat (Constructor; Auto).
Rewrite H2 in H6; Rewrite H1 in H8.
Inversion H6; Inversion H8; Auto.
(* st r1 (0) r5 *)
Apply I_st with Inv_gcd ([s:state] True)
 ([s,s':state] (EX va' | (EX vb' |
     (getValinH s' a)=(hvsome va') /\ (getValinH s' b)=(hvsome vb')
      /\ (gcd va' vb')=(gcd alpha beta)))); Auto.
Unfold Inv_gcd; Unfold getValinH; Unfold getValinR; Intros.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Unfold plus; Rewrite H6; Rewrite H8.
Constructor.
Apply indom_some with x1; Auto.
Intros; Repeat (Constructor; Auto).
Rewrite <- H9.
Exists (minus x x0); Exists x0; Repeat (Constructor; Auto).
Rewrite H0 in H3; Inversion H3; Rewrite H1 in H2; Inversion H2.
Rewrite H11 in H4; Rewrite H12 in H4; Rewrite <- H4; Apply gcd_minus; Auto.
(* yield *)
Apply I_yld with Inv_gcd A1_gcd G1_gcd; Auto.
Unfold Inv_gcd; Unfold getValinH; Intros; Induction s.
Inversion_clear H; Auto.
(* jd loop1 *)
Apply I_jd with CT1_gcd Inv_gcd a_loop1 g_loop1; Auto.
Unfold CT1_gcd; LookupCT.
Intros; Induction s.
Inversion_clear H; Unfold a_loop1; Constructor; Auto.
Save.

Lemma Infer_done1 : (Infer TSpec1_gcd a_done1 g_done1 I_done1).
Proof.
Unfold a_done1; Unfold g_done1; Unfold I_done1.
(* yield *)
Apply I_yld with Inv_gcd A1_gcd G1_gcd; Auto.
Unfold G1_gcd; Unfold A2_gcd; Unfold Inv_gcd; Unfold getValinH; Intros; Induction s.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Exists x; Exists x; Repeat (Constructor; Auto).
Intro; Absurd (gt x x); Auto with arith.
Unfold Inv_gcd; Unfold A1_gcd; Unfold getValinH; Intros; Induction s; Induction s'.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Exists x.
Rewrite H2 in H6; Rewrite H0 in H.
Inversion H6; Inversion H.
Rewrite H10 in H3; Rewrite H11 in H3.
Cut (hread a0 (1))=(Some nat x)/\(hread a0 (2))=(Some nat x); Auto.
Intro HH; Inversion_clear HH; Auto.
(* jd done1 *)
Apply I_jd with CT1_gcd Inv_gcd a_done1 g_done1; Auto.
Unfold CT1_gcd; LookupCT.
Intros; Induction s.
Inversion_clear H; Unfold a_done1; Constructor; Auto.
Save.

Lemma WF_CH1_gcd : (WF_CH TSpec1_gcd C1_gcd).
Proof.
Apply wf_c.
Replace (tgetCT TSpec1_gcd) with CT1_gcd;
 [Unfold CT1_gcd; Unfold C1_gcd | Auto].
Apply wf_c_cons.
Apply Infer_loop1.
Apply wf_c_cons.
Apply Infer_calc1.
Apply wf_c_cons.
Apply Infer_done1.
Apply wf_c_emp.
Save.

Lemma Infer_loop2 : (Infer TSpec2_gcd a_loop2 g_loop2 I_loop2).
Proof.
Unfold a_loop2; Unfold g_loop2; Unfold I_loop2.
(* yield *)
Apply I_yld with Inv_gcd A2_gcd G2_gcd; Auto.
Unfold G2_gcd; Unfold A1_gcd; Unfold Inv_gcd; Unfold getValinH; Intros; Induction s.
Inversion_clear H; SplitAllExs; SplitAllHyps.
Exists x; Exists x0; Repeat (Constructor; Auto).
Intro; Exists x; Exists x0; Repeat (Constructor; Auto).
(* movi r1 b *)
Apply I_seq with Inv_gcd ([s:state] (getValinR s r1)=b) G2_gcd; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Inversion_clear H0; Auto.
(* ld r2 r1 (0) *)
Apply I_ld with Inv_gcd
 ([s:state] (hvsome (getValinR s r2)) = (getValinH s b))
  G2_gcd; Auto.
Unfold Inv_gcd; Unfold getValinH; Unfold getValinR; Unfold plus.
Intros; SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H2; Clear H2; Exists x0; Repeat (Constructor; Auto).
Exists x; Exists x0; Auto.
(* movi r3 a *)
Apply I_seq with Inv_gcd
 ([s:state] (hvsome (getValinR s r2)) = (getValinH s b)
  /\ (getValinR s r3)=a)
  G2_gcd; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Inversion_clear H0; Auto.
(* ld r4 r3 (0) *)
Apply I_ld with Inv_gcd
 ([s:state] (hvsome (getValinR s r2)) = (getValinH s b)
  /\ (hvsome (getValinR s r4)) = (getValinH s a))
  G2_gcd; Auto.
Unfold Inv_gcd; Unfold getValinH; Unfold getValinR; Unfold plus.
Intros; SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H3; Clear H3; Exists x; Repeat (Constructor; Auto).
Exists x; Exists x0; Auto.
(* be r2 r4 done2 *)
Apply I_be with CT2_gcd Inv_gcd a_done2 ([s:state] True) g_done2 G2_gcd; Auto.
Unfold CT2_gcd; LookupCT.
Unfold Inv_gcd; Unfold a_done2; Unfold g_done2; Unfold getValinH; Unfold getValinR; Intros.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Constructor; Auto.
Rewrite H0 in H1; Rewrite H1 in H4; Rewrite H3 in H4; Exists x.
Repeat (Constructor; Auto).
Rewrite H2 in H4; Inversion H4.
Rewrite H7 in H6; Rewrite <- H6; Apply gcd_n_n.
(* yield *)
Apply I_yld with Inv_gcd A2_gcd G2_gcd; Auto.
Unfold Inv_gcd; Unfold G2_gcd; Unfold A1_gcd; Unfold getValinH; Intros; Induction s.
Inversion_clear H; SplitAllExs; SplitAllHyps.
Exists x; Exists x0; Repeat (Constructor; Auto).
Intro; Exists x; Exists x0; Repeat (Constructor; Auto).
(* movi r1 b *)
Apply I_seq with Inv_gcd ([s:state] (getValinR s r1)=b) G2_gcd; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Inversion_clear H0; Auto.
(* ld r2 r1 (0) *)
Apply I_ld with Inv_gcd
 ([s:state] (hvsome (getValinR s r2)) = (getValinH s b))
  G2_gcd; Auto.
Unfold Inv_gcd; Unfold getValinH; Unfold getValinR; Unfold plus.
Intros; SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H2; Clear H2; Exists x0; Repeat (Constructor; Auto).
Exists x; Exists x0; Auto.
(* movi r3 a *)
Apply I_seq with Inv_gcd
 ([s:state] (hvsome (getValinR s r2)) = (getValinH s b)
  /\ (getValinR s r3)=a)
  G2_gcd; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Inversion_clear H0; Auto.
(* ld r4 r3 (0) *)
Apply I_ld with Inv_gcd
 ([s:state] (hvsome (getValinR s r2)) = (getValinH s b)
  /\ (hvsome (getValinR s r4)) = (getValinH s a))
  G2_gcd; Auto.
Unfold Inv_gcd; Unfold getValinH; Unfold getValinR; Unfold plus.
Intros; SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H3; Clear H3; Exists x; Repeat (Constructor; Auto).
Exists x; Exists x0; Auto.
(* bgt r2 r4 calc2 *)
Apply I_bgt with CT2_gcd Inv_gcd a_calc2 ([s:state]True) g_calc2 G2_gcd; Auto.
Unfold CT2_gcd; LookupCT.
Unfold Inv_gcd; Unfold a_calc2; Unfold g_calc2; Unfold getValinH; Unfold getValinR; Intros.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Constructor; Auto.
Exists (lookupR R r2); Exists (lookupR R r4); Auto.
(* yield *)
Apply I_yld with Inv_gcd A2_gcd G2_gcd; Auto.
Unfold Inv_gcd; Unfold G2_gcd; Unfold A1_gcd; Unfold getValinH; Intros; Induction s.
Inversion_clear H; SplitAllExs; SplitAllHyps.
Exists x; Exists x0; Repeat (Constructor; Auto).
Intro; Exists x; Exists x0; Repeat (Constructor; Auto).
(* jd loop2 *)
Apply I_jd with CT2_gcd Inv_gcd a_loop2 g_loop2; Auto.
Unfold CT2_gcd; LookupCT.
Intros; Induction s.
Inversion_clear H; Unfold a_loop2; Constructor; Auto.
Save.

Lemma Infer_calc2 : (Infer TSpec2_gcd a_calc2 g_calc2 I_calc2).
Proof.
Unfold a_calc2; Unfold g_calc2; Unfold I_calc2.
(* yield *)
Apply I_yld with Inv_gcd A2_gcd G2_gcd; Auto.
Unfold G2_gcd; Unfold A1_gcd; Unfold Inv_gcd; Unfold getValinH; Intros; Induction s.
Inversion_clear H; SplitAllExs; SplitAllHyps.
Exists x0; Exists x; Repeat (Constructor; Auto).
Intro; Exists x1; Exists x2; Repeat (Constructor; Auto).
Unfold Inv_gcd; Unfold A2_gcd; Unfold getValinH; Intros; Induction s; Induction s'.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Clear H5 H4 H1 H7; Rewrite H in H6; Rewrite H0 in H2.
Exists x; Exists x0; Inversion H6; Inversion H2.
Rewrite H4 in H3; Rewrite H5 in H3.
Cut (hread a0 (2))=(Some nat x)/\(hread a0 (1))=(Some nat x0); Auto with arith.
Intro HH; Inversion_clear HH; Auto.
(* movi r1 b *)
Apply I_seq with Inv_gcd
 ([s:state] (EX va:nat | (EX vb:nat |
     (getValinH s a)=(hvsome va)
      /\ (getValinH s b)=(hvsome vb)
      /\ (gt vb va) /\ (getValinR s r1)=b)))
 ([s,s':state] (EX va':nat | (EX vb':nat |
     (getValinH s' a)=(hvsome va')
      /\ (getValinH s' b)=(hvsome vb')
      /\ (gcd va' vb')=(gcd alpha beta)))); Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Constructor; Auto.
Unfold getValinH in H0; Unfold getValinH in H1; Unfold getValinH; Constructor.
Exists x0; Exists x; Auto.
Intros; Induction s'; SplitAllExs; Unfold G2_gcd; Unfold A1_gcd.
Unfold getValinH; SplitAllHyps.
Exists x0; Exists x; Constructor; Auto.
Constructor; Auto.
Constructor; Intro.
Absurd (gt x x0); Auto with arith.
Exists x1; Exists x2; Repeat (Constructor; Auto).
(* ld r r1 (0) *)
Apply I_ld with Inv_gcd
 ([s:state] (EX va:nat | (EX vb:nat |
       (getValinH s a)=(hvsome va)
	/\ (getValinH s b)=(hvsome vb)
	/\ (gt vb va) /\ (getValinR s r1)=b
	/\ (hvsome (getValinR s r2))=(getValinH s b))))
  ([s,s':state] (EX va':nat | (EX vb':nat |
     (getValinH s' a)=(hvsome va')
      /\ (getValinH s' b)=(hvsome vb')
      /\ (gcd va' vb')=(gcd alpha beta)))); Auto.
Unfold Inv_gcd; Unfold getValinH; Unfold getValinR; Unfold plus.
Intros; SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H7; Exists x0; Repeat (Constructor; Auto).
Exists x1; Exists x2; Auto.
Exists x; Exists x0; Auto.
(* movi r3 a *)
Apply I_seq with Inv_gcd
 ([s:state] (EX va:nat | (EX vb:nat |
       (getValinH s a)=(hvsome va)
	/\ (getValinH s b)=(hvsome vb)
	/\ (gt vb va) /\ (getValinR s r1)=b
	/\ (hvsome (getValinR s r2))=(getValinH s b)
	/\(getValinR s r3)=a)))
 ([s,s':state] (EX va':nat | (EX vb':nat |
     (getValinH s' a)=(hvsome va')
      /\ (getValinH s' b)=(hvsome vb')
      /\ (gcd va' vb')=(gcd alpha beta)))); Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Repeat (Constructor; Auto).
Unfold getValinH in H0; Unfold getValinH in H1; Unfold getValinH.
Unfold getValinR in H3; Unfold getValinR.
Exists x; Exists x0; Repeat (Constructor; Auto).
(* ld r4 r3 (0) *)
Apply I_ld with Inv_gcd
 ([s:state] (EX va:nat | (EX vb:nat |
       (getValinH s a)=(hvsome va)
	/\ (getValinH s b)=(hvsome vb)
	/\ (gt vb va) /\ (getValinR s r1)=b
	/\ (hvsome (getValinR s r2))=(getValinH s b)
	/\(hvsome (getValinR s r4))=(getValinH s a))))
 ([s,s':state] (EX va':nat | (EX vb':nat |
     (getValinH s' a)=(hvsome va')
      /\ (getValinH s' b)=(hvsome vb')
      /\ (gcd va' vb')=(gcd alpha beta)))); Auto.
Unfold Inv_gcd; Unfold getValinH; Unfold getValinR; Unfold plus.
Intros; SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H9; Exists x; Repeat (Constructor; Auto).
Exists x1; Exists x2; Auto.
Exists x; Exists x0; Repeat (Constructor; Auto).
(* sub r5 r2 r4 *)
Apply I_seq with Inv_gcd
 ([s:state] (EX va:nat | (EX vb:nat |
       (getValinH s a)=(hvsome va)
	/\ (getValinH s b)=(hvsome vb)
	/\ (gt vb va) /\ (getValinR s r1)=b
	/\ (getValinR s r5) = (minus vb va))))
 ([s,s':state] (EX va':nat | (EX vb':nat |
     (getValinH s' a)=(hvsome va')
      /\ (getValinH s' b)=(hvsome vb')
      /\ (gcd va' vb')=(gcd alpha beta)))); Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_gcd in H0; Unfold Inv_gcd; Unfold getValinH in H0; Unfold getValinR in H0; Unfold getValinH; Unfold getValinR.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Repeat (Constructor; Auto).
Exists x1; Exists x2; Auto.
Exists x; Exists x0; Repeat (Constructor; Auto).
Rewrite H1 in H6; Rewrite H2 in H8.
Inversion H6; Inversion H8; Auto.
(* st r1 (0) r5 *)
Apply I_st with Inv_gcd ([s:state] True)
 ([s,s':state] (EX va' | (EX vb' |
     (getValinH s' a)=(hvsome va') /\ (getValinH s' b)=(hvsome vb')
      /\ (gcd va' vb')=(gcd alpha beta)))); Auto.
Unfold Inv_gcd; Unfold getValinH; Unfold getValinR; Intros.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Unfold plus; Rewrite H6; Rewrite H8.
Constructor.
Apply indom_some with x2; Auto.
Intros; Repeat (Constructor; Auto).
Rewrite <- H9.
Exists x; Exists (minus x0 x); Repeat (Constructor; Auto).
Rewrite H0 in H3; Inversion H3; Rewrite H1 in H2; Inversion H2.
Rewrite H11 in H4; Rewrite H12 in H4; Rewrite <- H4.
Replace (gcd x (minus x0 x)) with (gcd (minus x0 x) x);
 [Idtac | Apply gcd_sym].
Replace (gcd x x0) with (gcd x0 x); [Apply gcd_minus; Auto | Apply gcd_sym].
(* yield *)
Apply I_yld with Inv_gcd A2_gcd G2_gcd; Auto.
Unfold Inv_gcd; Unfold getValinH; Intros; Induction s.
Inversion_clear H; Auto.
(* jd loop2 *)
Apply I_jd with CT2_gcd Inv_gcd a_loop2 g_loop2; Auto.
Unfold CT2_gcd; LookupCT.
Intros; Induction s.
Inversion_clear H; Unfold a_loop2; Constructor; Auto.
Save.

Lemma Infer_done2 : (Infer TSpec2_gcd a_done2 g_done2 I_done2).
Proof.
Unfold a_done2; Unfold g_done2; Unfold I_done2.
(* yield *)
Apply I_yld with Inv_gcd A2_gcd G2_gcd; Auto.
Unfold G2_gcd; Unfold A1_gcd; Unfold Inv_gcd; Unfold getValinH; Intros; Induction s.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Exists x; Exists x; Repeat (Constructor; Auto).
Intro; Absurd (gt x x); Auto with arith.
Unfold Inv_gcd; Unfold A2_gcd; Unfold getValinH; Intros; Induction s; Induction s'.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Exists x.
Rewrite H2 in H6; Rewrite H0 in H.
Inversion H6; Inversion H.
Rewrite H10 in H3; Rewrite H11 in H3.
Cut (hread a0 (2))=(Some nat x)/\(hread a0 (1))=(Some nat x); Auto.
Intro HH; Inversion_clear HH; Auto.
(* jd done2 *)
Apply I_jd with CT2_gcd Inv_gcd a_done2 g_done2; Auto.
Unfold CT2_gcd; LookupCT.
Intros; Induction s.
Inversion_clear H; Unfold a_done2; Constructor; Auto.
Save.

Lemma WF_CH2_gcd : (WF_CH TSpec2_gcd C2_gcd).
Proof.
Apply wf_c.
Replace (tgetCT TSpec2_gcd) with CT2_gcd;
 [Unfold CT2_gcd; Unfold C2_gcd | Auto].
Apply wf_c_cons.
Apply Infer_loop2.
Apply wf_c_cons.
Apply Infer_calc2.
Apply wf_c_cons.
Apply Infer_done2.
Apply wf_c_emp.
Save.

Theorem WF_gcd : (WFW Spec_gcd p1_gcd p2_gcd g1_gcd Prog_gcd).
Proof.
Unfold Prog_gcd.
Apply wf1_w with TSpec1_gcd TSpec2_gcd Inv_gcd A2_gcd G2_gcd; Auto.
Apply WF_CH1_gcd.
Apply WF_CH2_gcd.
Unfold Inv_gcd;Unfold S_gcd; Unfold a_loop1; Constructor; Auto.
Unfold H_gcd; Exists alpha; Exists beta; Auto.
Unfold a_loop2; Auto.
Unfold a_loop2; Auto.
Apply I_jd with CT1_gcd Inv_gcd ([s:state] True) G1_gcd; Auto.
Unfold CT1_gcd; LookupCT.
Apply I_jd with CT2_gcd Inv_gcd ([s:state] True) G2_gcd; Auto.
Unfold CT2_gcd; LookupCT.
Save.
