(****** CCAP Implementation for Dining Philsophers ******)
(****** created Mar. 29 2004, revised Jun. 29 2004 ******)

Require Arith.
Require ccap.

Syntactic Definition fork1 := (1).
Syntactic Definition fork2 := (2).

Definition H_philo : heap :=
 [x:addr]
 (if (beq_nat x fork1) then (hvsome O) else
   (if (beq_nat x fork2) then (hvsome O) else hvnone)).
Definition S_philo : state := (H_philo, emptyR).

Syntactic Definition fa1 := (1).
Syntactic Definition fb1 := (2).
Syntactic Definition fa2 := (3).
Syntactic Definition fb2 := (4).

Definition I_fa1 : iseq :=
 (yield
   (seq (movi r1 fork1)
     (ld r2 r1 O
       (seq (movi r3 O)
	 (bgt r2 r3 fa1
	   (seq (movi r2 (1))
	     (st r1 O r2
	       (yield
		 (jd fb1))))))))).
Definition I_fb1 : iseq :=
 (yield
   (seq (movi r1 fork2)
     (ld r2 r1 O
       (seq (movi r3 O)
	 (bgt r2 r3 fb1
	   (seq (movi r2 (1))
	     (st r1 O r2
	       (yield
		 (seq (movi r1 fork2)
		   (seq (movi r2 O)
		     (st r1 O r2
		       (yield
			 (seq (movi r1 fork1)
			   (seq (movi r2 O)
			     (st r1 O r2
			       (yield
				 (jd fa1))))))))))))))))).
Definition C1_philo : codeheap :=
 (consC fa1 I_fa1
   (consC fb1 I_fb1 emptyC)).

Definition I_fa2 : iseq :=
 (yield
   (seq (movi r1 fork1)
     (ld r2 r1 O
       (seq (movi r3 O)
	 (bgt r2 r3 fa2
	   (seq (movi r2 (2))
	     (st r1 O r2
	       (yield
		 (jd fb2))))))))).
Definition I_fb2 : iseq :=
 (yield
   (seq (movi r1 fork2)
     (ld r2 r1 O
       (seq (movi r3 O)
	 (bgt r2 r3 fb2
	   (seq (movi r2 (2))
	     (st r1 O r2
	       (yield
		 (seq (movi r1 fork2)
		   (seq (movi r2 O)
		     (st r1 O r2
		       (yield
			 (seq (movi r1 fork1)
			   (seq (movi r2 O)
			     (st r1 O r2
			       (yield
				 (jd fa2))))))))))))))))).
Definition C2_philo : codeheap :=
 (consC fa2 I_fa2
   (consC fb2 I_fb2 emptyC)).

Definition Prog_philo : prog :=
 ((S_philo, (C1_philo, C2_philo)), (((jd fa1), (jd fa2)), (1))).

Definition Inv_philo : assert :=
 [s:state]
  let (H, R) = s in (hindom fork1 H)
  /\((hread H fork2) = (hvsome O) \/ (hread H fork2) = (hread H fork1)).

Definition A1_philo : guar :=
 [s,s':state]
  let (H, R) =s in let (H', R') = s' in
 (hindom fork1 H') /\ (hindom fork2 H')
  /\ ((hread H' fork2) = (hvsome (2)) -> (hread H' fork1) = (hvsome (2)))
  /\ ((hread H fork1) = (hvsome (1)) -> (hread H' fork1) = (hvsome (1)))
  /\ ((hread H' fork1) = (hvsome (1)) -> (hread H fork1) = (hvsome (1)))
  /\ ((hread H fork2) = (hvsome (1)) -> (hread H' fork2) = (hvsome (1)))
  /\ ((hread H' fork2) = (hvsome (1)) -> (hread H fork2) = (hvsome (1))).
Definition A2_philo : guar :=
 [s,s':state]
  let (H, R) =s in let (H', R') = s' in
 (hindom fork1 H') /\ (hindom fork2 H')
  /\ ((hread H' fork2) = (hvsome (1)) -> (hread H' fork1) = (hvsome (1)))
  /\ ((hread H fork1) = (hvsome (2)) -> (hread H' fork1) = (hvsome (2)))
  /\ ((hread H' fork1) = (hvsome (2)) -> (hread H fork1) = (hvsome (2)))
  /\ ((hread H fork2) = (hvsome (2)) -> (hread H' fork2) = (hvsome (2)))
  /\ ((hread H' fork2) = (hvsome (2)) -> (hread H fork2) = (hvsome (2))).
Definition G1_philo : guar := A2_philo.
Definition G2_philo : guar := A1_philo.

Definition a_fa1 : assert :=
 [s:state] (EX vf1 | (EX vf2 |
     (getValinH s fork1) = (hvsome vf1)
      /\ (getValinH s fork2) = (hvsome vf2)
      /\ ~(vf1 = (1)) /\ ~(vf2 = (1)))).
Definition g_fa1 : guar := G1_philo.
Definition a_fb1 : assert :=
 [s:state] (EX vf2 |
     (getValinH s fork1) = (hvsome (1))
      /\ (getValinH s fork2) = (hvsome vf2)
      /\ ~(vf2 = (1))).
Definition g_fb1 : guar := G1_philo.
Definition CT1_philo : codeheapty :=
 (consCT fa1 a_fa1 g_fa1
   (consCT fb1 a_fb1 g_fb1 emptyCT)).

Definition a_fa2 : assert :=
 [s:state] (EX vf1 | (EX vf2 |
     (getValinH s fork1) = (hvsome vf1)
      /\ (getValinH s fork2) = (hvsome vf2)
      /\ ~(vf1 = (2)) /\ ~(vf2 = (2)))).
Definition g_fa2 : guar := G2_philo.
Definition a_fb2 : assert :=
 [s:state] (EX vf2 |
     (getValinH s fork1) = (hvsome (2))
      /\ (getValinH s fork2) = (hvsome vf2)
      /\ ~(vf2 = (2))).
Definition g_fb2 : guar := G2_philo.
Definition CT2_philo : codeheapty :=
 (consCT fa2 a_fa2 g_fa2
   (consCT fb2 a_fb2 g_fb2 emptyCT)).

Definition Spec_philo :=
 (pairT
   (pairT Inv_philo (pairT CT1_philo CT2_philo))
   (pairT (pairT A1_philo G1_philo) (pairT A2_philo G2_philo))).

Definition TSpec1_philo :=
 (pairT
   (pairT Inv_philo CT1_philo)
   (pairT A1_philo G1_philo)).
Definition TSpec2_philo :=
 (pairT
   (pairT Inv_philo CT2_philo)
   (pairT A2_philo G2_philo)).

Syntactic Definition p1_philo := a_fa1.
Syntactic Definition p2_philo := a_fa2.
Syntactic Definition g1_philo := G1_philo.

Lemma Infer_fa1 : (Infer TSpec1_philo a_fa1 g_fa1 I_fa1).
Proof.
Unfold a_fa1; Unfold g_fa1; Unfold I_fa1.
(* yield *)
Apply I_yld with Inv_philo A1_philo G1_philo; Auto.
Unfold G1_philo; Unfold getValinH; Intros; Induction s; Unfold A2_philo.
Unfold Inv_philo in H; SplitAllHyps.
SplitAllExs; SplitAllHyps; Repeat (Constructor; Auto).
Apply indom_some with x0; Auto.
Rewrite H1; Intro EQ; Inversion EQ; Absurd x0=(1); Auto.
Unfold getValinH; Intros; Induction s; Induction s'.
Unfold Inv_philo in H; Unfold A1_philo in H0; SplitAllHyps.
SplitAllExs; SplitAllHyps.
Rewrite H6 in H3; Rewrite H6 in H4; Rewrite H8 in H5; Rewrite H8 in H7.
Inversion_clear H1; Inversion_clear H0.
Unfold hread in H4; Unfold hread in H7; Unfold hread.
Rewrite H11 in H4; Rewrite H1 in H7; Rewrite H11; Rewrite H1.
Exists n; Exists n0; Clear H1 H11 H3 H5 H8 H6 H9 H H2 a a0.
Repeat (Constructor; Auto).
Intro; Absurd (Some nat x)=(Some nat (1)); Auto.
Intro EQ; Inversion EQ; Absurd x=(1); Auto.
Intro; Absurd (Some nat x0)=(Some nat (1)); Auto.
Intro EQ; Inversion EQ; Absurd x0=(1); Auto.
(* movi r1 fork1 *)
Apply I_seq with Inv_philo
 ([s:state] (a_fa1 s) /\ (getValinR s r1)=fork1)
  G1_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; Constructor; Auto.
(* ld r2 r1 (0) *)
Apply I_ld with Inv_philo
 ([s:state] (a_fa1 s) /\ (getValinR s r1)=fork1
  /\ (hvsome (getValinR s r2))=(getValinH s fork1))
  G1_philo; Auto.
Unfold getValinR; Unfold getValinH; Unfold Inv_philo; Intros.
SplitAllHyps.
Inversion H2; Rewrite H3; Unfold plus.
Exists n; Repeat (Constructor; Auto).
(* movi r3 0 *)
Apply I_seq with Inv_philo
 ([s:state] (a_fa1 s) /\ (getValinR s r1)=fork1
  /\ (hvsome (getValinR s r2))=(getValinH s fork1) /\ (getValinR s r3)=O)
  G1_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; Constructor; Auto.
(* bgt r2 r3 fa1 *)
Apply I_bgt with CT1_philo Inv_philo a_fa1
 ([s:state] (EX vf2 | (getValinH s fork1)=(hvsome O)
  /\ (getValinH s fork2)=(hvsome vf2) /\ ~vf2=(1) /\ (getValinR s r1)=fork1))
  g_fa1 G1_philo; Auto.
Unfold CT1_philo; LookupCT.
Unfold getValinR; Unfold getValinH; Unfold a_fa1; Unfold Inv_philo; Unfold g_fa1; Unfold G1_philo; Intros.
SplitAllHyps; Constructor; Auto.
Unfold Inv_philo; Unfold a_fa1; Unfold getValinH; Unfold getValinR; Intros.
SplitAllHyps; Constructor; Auto.
SplitAllExs; SplitAllHyps.
Exists x0; Constructor; Auto.
Rewrite H6 in H0; Rewrite <- H4.
Replace (lookupR R r2) with O; Auto with arith.
(* movi r2 1 *)
Apply I_seq with Inv_philo
 ([s:state] (EX vf2 | (getValinH s fork1)=(hvsome O)
  /\ (getValinH s fork2)=(hvsome vf2) /\ ~vf2=(1)
  /\ (getValinR s r1)=fork1 /\ (getValinR s r2)=(1)))
 ([s,s':state] let (H, R) = s in let (H', R') = s' in
   (hindom fork1 H') /\ (hindom fork2 H')
    /\ ((hread H' fork2) = (hvsome (1)) -> (hread H' fork1) = (hvsome (1)))
    /\ ~(hread H' fork1)=(hvsome (2))
    /\ ((hread H fork2) = (hvsome (2)) -> (hread H' fork2) = (hvsome (2)))
    /\ ((hread H' fork2) = (hvsome (2)) -> (hread H fork2) = (hvsome (2))));
    Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; Constructor; Auto.
Unfold getValinH in H1; Unfold getValinH; Unfold getValinR.
Inversion_clear H1; SplitAllHyps.
Constructor.
Exists x; Constructor; Auto.
Intros; Induction s'.
Unfold G1_philo; Unfold A2_philo.
SplitAllHyps; Repeat (Constructor; Auto).
Intro EQ; Rewrite EQ in H1; Inversion H1.
Intro; Absurd (hread a1 (1))=(Some nat (2)); Auto.
(* st r1 (0) r2 *)
Apply I_st with Inv_philo
 ([s:state] (EX vf2 | (getValinH s fork1)=(hvsome (1))
  /\ (getValinH s fork2)=(hvsome vf2) /\ ~vf2=(1)))
 ([s,s':state] let (H, R) = s in let (H', R') = s' in
   (hindom fork1 H') /\ (hindom fork2 H')
    /\ ((hread H' fork2) = (hvsome (1)) -> (hread H' fork1) = (hvsome (1)))
    /\ ~(hread H' fork1)=(hvsome (2))
    /\ ((hread H fork2) = (hvsome (2)) -> (hread H' fork2) = (hvsome (2)))
    /\ ((hread H' fork2) = (hvsome (2)) -> (hread H fork2) = (hvsome (2))));
    Auto.
Unfold getValinR; Unfold getValinH; Unfold Inv_philo; Intros.
Unfold plus; SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H5; Rewrite H7; Clear H5; Clear H7; Constructor; Auto.
Intros.
Cut (hread H5 (2))=(hread H (2));
 [Intro EQ; Rewrite EQ | Rewrite <- H6; Apply local_write; Auto].
Constructor; Auto.
Cut (hread H5 (1)) = (hvsome (1));
 [Intro EQ1; Rewrite EQ1 | Rewrite <- H6; Apply write_read].
Constructor; [Constructor | Exists x; Auto].
Apply indom_some with (1); Auto.
Left.
Inversion_clear H3; Auto.
Rewrite H7; Auto.
(* yield *)
Apply I_yld with Inv_philo A1_philo G1_philo; Auto.
Unfold Inv_philo; Unfold getValinH; Intros; Induction s.
SplitAllHyps; SplitAllExs; SplitAllHyps; Repeat (Constructor; Auto).
Apply indom_some with x; Auto.
Intro EQ; Rewrite EQ in H1; Inversion H1.
Unfold getValinH; Intros; Induction s; Induction s'.
Unfold Inv_philo in H; Unfold A1_philo in H0; SplitAllHyps.
SplitAllExs; SplitAllHyps.
Rewrite H8 in H3; Rewrite H8 in H4; Rewrite H6 in H5; Rewrite H6 in H7.
Inversion_clear H1; Inversion_clear H0.
Exists n0; Repeat (Constructor; Auto).
Unfold hread in H7; Rewrite H1 in H7.
Intro; Absurd (Some nat x)=(Some nat (1)); Auto.
Intro EQ; Inversion EQ; Absurd x=(1); Auto.
(* jd fb1 *)
Apply I_jd with CT1_philo Inv_philo a_fb1 g_fb1; Auto.
Unfold CT1_philo; LookupCT.
Intros; Induction s; Unfold a_fb1; SplitAllHyps; Auto.
Save.

Lemma Infer_fb1 : (Infer TSpec1_philo a_fb1 g_fb1 I_fb1).
Proof.
Unfold a_fb1; Unfold g_fb1; Unfold I_fb1.
(* yield *)
Apply I_yld with Inv_philo A1_philo G1_philo; Auto.
Unfold G1_philo; Unfold getValinH; Intros; Induction s; Unfold A2_philo.
Unfold Inv_philo in H; SplitAllHyps.
SplitAllExs; SplitAllHyps; Repeat (Constructor; Auto).
Apply indom_some with x; Auto.
Unfold getValinH; Intros; Induction s; Induction s'.
Unfold Inv_philo in H; Unfold A1_philo in H0; SplitAllHyps.
SplitAllExs; SplitAllHyps.
Rewrite H8 in H3; Rewrite H8 in H4; Rewrite H6 in H5; Rewrite H6 in H7.
Inversion_clear H1; Inversion_clear H0.
Exists n0; Repeat (Constructor; Auto).
Unfold hread in H7; Rewrite H1 in H7.
Intro; Absurd (Some nat x)=(Some nat (1)); Auto.
Intro EQ; Inversion EQ; Absurd x=(1); Auto.
(* movi r1 fork2 *)
Apply I_seq with Inv_philo
 ([s:state] (a_fb1 s) /\ (getValinR s r1)=fork2)
  G1_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; Constructor; Auto.
(* ld r2 r1 (0) *)
Apply I_ld with Inv_philo
 ([s:state] (a_fb1 s) /\ (getValinR s r1)=fork2
  /\ (hvsome (getValinR s r2))=(getValinH s fork2))
  G1_philo; Auto.
Unfold getValinR; Unfold getValinH; Unfold Inv_philo; Intros.
SplitAllHyps.
Cut (hindom (2) H).
Intro HH; Inversion HH; Rewrite H3; Unfold plus.
Exists n; Repeat (Constructor; Auto).
Inversion_clear H2.
Inversion_clear H4; [Apply indom_some with O | Apply indom_some with n]; Auto.
Unfold hread in H2; Rewrite <- H1; Auto.
(* movi r3 (0) *)
Apply I_seq with Inv_philo
 ([s:state] (a_fb1 s) /\ (getValinR s r1)=fork2
  /\ (hvsome (getValinR s r2))=(getValinH s fork2) /\ (getValinR s r3)=O)
  G1_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; Constructor; Auto.
(* bgt r2 r3 fb1 *)
Apply I_bgt with CT1_philo Inv_philo a_fb1
 ([s:state] (getValinH s fork1)=(hvsome (1))
  /\ (getValinH s fork2)=(hvsome O) /\ (getValinR s r1)=fork2)
  g_fb1 G1_philo; Auto.
Unfold CT1_philo; LookupCT.
Unfold getValinR; Unfold getValinH; Unfold a_fb1; Unfold Inv_philo; Unfold g_fb1; Unfold G1_philo; Intros.
SplitAllHyps; Constructor; Auto.
Unfold Inv_philo; Unfold a_fb1; Unfold getValinH; Unfold getValinR; Intros.
SplitAllHyps; Constructor; Auto.
SplitAllExs; SplitAllHyps; Repeat (Constructor; Auto).
Rewrite H6 in H0; Rewrite <- H4.
Replace (lookupR R r2) with O; Auto with arith.
(* movi r2 1 *)
Apply I_seq with Inv_philo
 ([s:state] (getValinH s fork1)=(hvsome (1))
  /\ (getValinH s fork2)=(hvsome O)
  /\ (getValinR s r1)=fork2 /\ (getValinR s r2)=(1))
 ([s,s':state] let (H, R) = s in let (H', R') = s' in
   (hindom fork1 H') /\ (hindom fork2 H')
    /\ ((hread H' fork2) = (hvsome (1)) -> (hread H' fork1) = (hvsome (1)))
    /\ ~(hread H' fork1)=(hvsome (2))
    /\ ~(hread H' fork2)=(hvsome (2)));
    Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo.
Unfold getValinH in H0; Unfold getValinH; Unfold getValinR in H0; Unfold getValinR; SplitAllHyps.
Repeat (Constructor; Auto).
Intros; Induction s'.
Unfold G1_philo; Unfold A2_philo; SplitAllHyps.
Repeat (Constructor; Auto).
Intro EQ; Rewrite EQ in H0; Inversion H0.
Intro; Absurd (hread a1 (1))=(Some nat (2)); Auto.
Intro EQ; Rewrite EQ in H1; Inversion H1.
Intro; Absurd (hread a1 (2))=(Some nat (2)); Auto.
(* st r1 (0) r2 *)
Apply I_st with Inv_philo
 ([s:state] (getValinH s fork1)=(hvsome (1))
  /\ (getValinH s fork2)=(hvsome (1)))
 ([s,s':state] let (H, R) = s in let (H', R') = s' in
   (hindom fork1 H') /\ (hindom fork2 H')
    /\ ((hread H' fork2) = (hvsome (1)) -> (hread H' fork1) = (hvsome (1)))
    /\ ~(hread H' fork1)=(hvsome (2))
    /\ ~(hread H' fork2)=(hvsome (2)));
    Auto.
Unfold getValinR; Unfold getValinH; Unfold Inv_philo; Intros.
Unfold plus; SplitAllHyps.
Rewrite H3; Rewrite H5; Clear H3; Clear H5; Constructor.
Apply indom_some with O; Auto.
Intros.
Cut (hread H1 (1))=(hread H (1));
 [Intro EQ; Rewrite EQ | Rewrite <- H3; Apply local_write; Auto].
Constructor; Auto.
Cut (hread H1 (2)) = (hvsome (1));
 [Intro EQ1; Rewrite EQ1 | Rewrite <- H3; Apply write_read].
Repeat (Constructor; Auto).
Rewrite H0 in EQ; Apply indom_some with (1); Auto.
(* yield *)
Apply I_yld with Inv_philo A1_philo G1_philo; Auto.
Unfold Inv_philo; Unfold getValinH; Intros; Induction s.
SplitAllHyps; Repeat (Constructor; Auto).
Apply indom_some with (1); Auto.
Intro EQ; Rewrite EQ in H; Inversion H.
Intro EQ; Rewrite EQ in H2; Inversion H2.
Unfold getValinH; Intros; Induction s; Induction s'.
Unfold Inv_philo in H; Unfold A1_philo in H0; SplitAllHyps.
Constructor; Auto.
(* movi r1 fork2 *)
Apply I_seq with Inv_philo
 ([s:state] (getValinH s fork1)=(hvsome (1))
  /\ (getValinH s fork2)=(hvsome (1)) /\ (getValinR s r1)=fork2)
  G1_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; Constructor; Auto.
(* movi r2 0 *)
Apply I_seq with Inv_philo
 ([s:state] (getValinH s fork1)=(hvsome (1))
  /\ (getValinH s fork2)=(hvsome (1)) /\ (getValinR s r1)=fork2
  /\ (getValinR s r2)=O)
  G1_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; Constructor; Auto.
(* st r1 (0) r2 *)
Apply I_st with Inv_philo
 ([s:state] (EX vf2 | (getValinH s fork1)=(hvsome (1))
  /\ (getValinH s fork2)=(hvsome vf2) /\ ~vf2=(1)))
  G1_philo; Auto.
Unfold getValinH; Unfold getValinR; Unfold Inv_philo; Unfold plus; Intros.
SplitAllHyps.
Rewrite H3; Rewrite H5; Clear H3; Clear H5; Constructor.
Apply indom_some with (1); Auto.
Intros.
Cut (hread H1 (2))=(hvsome O);
 [Intro EQ; Rewrite EQ | Rewrite <- H3; Apply write_read].
Cut (hread H1 (1))=(hread H (1));
 [Intro EQ1; Rewrite EQ1 | Rewrite <- H3; Apply local_write; Auto].
Repeat (Constructor; Auto).
Rewrite H0 in EQ1; Apply indom_some with (1); Auto.
Exists O; Auto.
Unfold G1_philo; Unfold A2_philo.
Intros; Induction s.
SplitAllHyps; Repeat (Constructor; Auto).
Intro FF; Rewrite FF in H0; Inversion H0.
Intro; Cut (hread H1 (1))=(Some nat (2)); Auto.
Intro FF; Rewrite FF in EQ1; Rewrite <- EQ1 in H0; Inversion H0.
Intro FF; Rewrite FF in H2; Inversion H2.
Intro; Cut (hread H1 (2))=(Some nat (2)); Auto.
Intro FF; Rewrite FF in EQ; Inversion EQ.
(* yield *)
Apply I_yld with Inv_philo A1_philo G1_philo; Auto.
Unfold Inv_philo; Unfold getValinH; Intros; Induction s.
Unfold G1_philo; Unfold A2_philo; SplitAllHyps; Repeat (Constructor; Auto).
Inversion_clear H2.
Apply indom_some with O; Auto.
Inversion_clear H.
Apply indom_some with n.
Rewrite <- H2; Unfold hread in H0; Auto.
Inversion_clear H2.
Intro EQ; Rewrite EQ in H0; Inversion H0.
Intro EQ; Rewrite <- EQ; Auto.
Unfold getValinH; Intros; Induction s; Induction s'.
Unfold Inv_philo in H; Unfold A1_philo in H0; SplitAllHyps.
Inversion_clear H8; SplitAllHyps.
Inversion_clear H0.
Exists n; Repeat (Constructor; Auto).
Intro EQ; Rewrite EQ in H10.
Absurd (hread a (2))=(Some nat (1)); Auto.
Rewrite H6; Intro EQ1; Inversion EQ1; Absurd x=(1); Auto.
(* movi r1 fork1 *)
Apply I_seq with Inv_philo
 ([s:state] (EX vf2 | (getValinH s fork1)=(hvsome (1))
  /\ (getValinH s fork2)=(hvsome vf2) /\ ~vf2=(1) /\ (getValinR s r1)=fork1))
  G1_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; SplitAllExs.
SplitAllHyps; Repeat (Constructor; Auto).
Exists x; Constructor; Auto.
(* movi r2 0 *)
Apply I_seq with Inv_philo
 ([s:state] (EX vf2 | (getValinH s fork1)=(hvsome (1))
  /\ (getValinH s fork2)=(hvsome vf2) /\ ~vf2=(1) /\ (getValinR s r1)=fork1
  /\ (getValinR s r2)=O))
  G1_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; SplitAllExs.
SplitAllHyps; Repeat (Constructor; Auto).
Exists x; Constructor; Auto.
(* st r1 (0) r2 *)
Apply I_st with Inv_philo
 ([s:state] (EX vf1 | (EX vf2 | (getValinH s fork1)=(hvsome vf1)
  /\ (getValinH s fork2)=(hvsome vf2) /\ ~vf1=(1)/\ ~vf2=(1))))
  G1_philo; Auto.
Unfold getValinH; Unfold getValinR; Unfold Inv_philo; Unfold plus; Intros.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H5; Rewrite H7; Clear H5; Clear H7; Constructor; Auto.
Intros.
Cut (hread H5 (2))=(hread H (2));
 [Intro EQ; Rewrite EQ | Rewrite <- H6; Apply local_write; Auto].
Cut (hread H5 (1))=(hvsome O);
 [Intro EQ1; Rewrite EQ1 | Rewrite <- H6; Apply write_read].
Repeat (Constructor; Auto).
Apply indom_some with O; Auto.
Inversion_clear H3; Auto.
Rewrite H7 in H1; Rewrite H2 in H1; Inversion H1; Absurd x=(1); Auto.
Exists O; Exists x; Auto.
Unfold G1_philo; Unfold A2_philo; Intros; Induction s.
SplitAllHyps; Repeat (Constructor; Auto).
Intro FF; Rewrite FF in H2; Inversion H2.
Intro; Cut (hread H5 (1))=(Some nat (2)); Auto.
Intro FF; Rewrite FF in EQ1; Inversion EQ1.
Rewrite <- EQ; Auto.
Rewrite <- EQ; Auto.
(* yield *)
Apply I_yld with Inv_philo A1_philo G1_philo; Auto.
Unfold Inv_philo; Unfold getValinH; Intros; Induction s.
Unfold G1_philo; Unfold A2_philo; SplitAllHyps; SplitAllExs; SplitAllHyps.
Repeat (Constructor; Auto).
Apply indom_some with x0; Auto.
Intro EQ; Rewrite EQ in H1; Inversion H1; Absurd x0=(1); Auto.
Unfold getValinH; Intros; Induction s; Induction s'.
Unfold Inv_philo in H; Unfold A1_philo in H0; SplitAllHyps.
SplitAllExs; SplitAllHyps.
Rewrite H6 in H3; Rewrite H6 in H4; Rewrite H8 in H5; Rewrite H8 in H7.
Inversion_clear H1; Inversion_clear H0.
Exists n; Exists n0; Repeat (Constructor; Auto).
Unfold hread in H4; Rewrite H11 in H4.
Intro; Absurd (Some nat x)=(Some nat (1)); Auto.
Intro EQ; Inversion EQ; Absurd x=(1); Auto.
Unfold hread in H7; Rewrite H1 in H7.
Intro; Absurd (Some nat x0)=(Some nat (1)); Auto.
Intro EQ; Inversion EQ; Absurd x0=(1); Auto.
(* jd fa1 *)
Apply I_jd with CT1_philo Inv_philo a_fa1 g_fa1; Auto.
Unfold CT1_philo; LookupCT.
Intros; Induction s; Unfold a_fa1; SplitAllHyps; Auto.
Save.

Lemma WF_CH1_philo : (WF_CH TSpec1_philo C1_philo).
Proof.
Apply wf_c.
Replace (tgetCT TSpec1_philo) with CT1_philo;
 [Unfold CT1_philo; Unfold C1_philo | Auto].
Apply wf_c_cons.
Apply Infer_fa1.
Apply wf_c_cons.
Apply Infer_fb1.
Apply wf_c_emp.
Save.

Lemma Infer_fa2 : (Infer TSpec2_philo a_fa2 g_fa2 I_fa2).
Proof.
Unfold a_fa2; Unfold g_fa2; Unfold I_fa2.
(* yield *)
Apply I_yld with Inv_philo A2_philo G2_philo; Auto.
Unfold G2_philo; Unfold getValinH; Intros; Induction s; Unfold A1_philo.
Unfold Inv_philo in H; SplitAllHyps.
SplitAllExs; SplitAllHyps; Repeat (Constructor; Auto).
Apply indom_some with x0; Auto.
Rewrite H1; Intro EQ; Inversion EQ; Absurd x0=(2); Auto.
Unfold getValinH; Intros; Induction s; Induction s'.
Unfold Inv_philo in H; Unfold A2_philo in H0; SplitAllHyps.
SplitAllExs; SplitAllHyps.
Rewrite H6 in H3; Rewrite H6 in H4; Rewrite H8 in H5; Rewrite H8 in H7.
Inversion_clear H1; Inversion_clear H0.
Unfold hread in H4; Unfold hread in H7; Unfold hread.
Rewrite H11 in H4; Rewrite H1 in H7; Rewrite H11; Rewrite H1.
Exists n; Exists n0; Clear H1 H11 H3 H5 H8 H6 H9 H H2 a a0.
Repeat (Constructor; Auto).
Intro; Absurd (Some nat x)=(Some nat (2)); Auto.
Intro EQ; Inversion EQ; Absurd x=(2); Auto.
Intro; Absurd (Some nat x0)=(Some nat (2)); Auto.
Intro EQ; Inversion EQ; Absurd x0=(2); Auto.
(* movi r1 fork1 *)
Apply I_seq with Inv_philo
 ([s:state] (a_fa2 s) /\ (getValinR s r1)=fork1)
  G2_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; Constructor; Auto.
(* ld r2 r1 (0) *)
Apply I_ld with Inv_philo
 ([s:state] (a_fa2 s) /\ (getValinR s r1)=fork1
  /\ (hvsome (getValinR s r2))=(getValinH s fork1))
  G2_philo; Auto.
Unfold getValinR; Unfold getValinH; Unfold Inv_philo; Intros.
SplitAllHyps.
Inversion H2; Rewrite H3; Unfold plus.
Exists n; Repeat (Constructor; Auto).
(* movi r3 0 *)
Apply I_seq with Inv_philo
 ([s:state] (a_fa2 s) /\ (getValinR s r1)=fork1
  /\ (hvsome (getValinR s r2))=(getValinH s fork1) /\ (getValinR s r3)=O)
  G2_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; Constructor; Auto.
(* bgt r2 r3 fa2 *)
Apply I_bgt with CT2_philo Inv_philo a_fa2
 ([s:state] (EX vf2 | (getValinH s fork1)=(hvsome O)
  /\ (getValinH s fork2)=(hvsome vf2) /\ ~vf2=(2) /\ (getValinR s r1)=fork1))
  g_fa2 G2_philo; Auto.
Unfold CT2_philo; LookupCT.
Unfold getValinR; Unfold getValinH; Unfold a_fa2; Unfold Inv_philo; Unfold g_fa2; Unfold G2_philo; Intros.
SplitAllHyps; Constructor; Auto.
Unfold Inv_philo; Unfold a_fa2; Unfold getValinH; Unfold getValinR; Intros.
SplitAllHyps; Constructor; Auto.
SplitAllExs; SplitAllHyps.
Exists x0; Constructor; Auto.
Rewrite H6 in H0; Rewrite <- H4.
Replace (lookupR R r2) with O; Auto with arith.
(* movi r2 2 *)
Apply I_seq with Inv_philo
 ([s:state] (EX vf2 | (getValinH s fork1)=(hvsome O)
  /\ (getValinH s fork2)=(hvsome vf2) /\ ~vf2=(2)
  /\ (getValinR s r1)=fork1 /\ (getValinR s r2)=(2)))
 ([s,s':state] let (H, R) = s in let (H', R') = s' in
   (hindom fork1 H') /\ (hindom fork2 H')
    /\ ((hread H' fork2) = (hvsome (2)) -> (hread H' fork1) = (hvsome (2)))
    /\ ~(hread H' fork1)=(hvsome (1))
    /\ ((hread H fork2) = (hvsome (1)) -> (hread H' fork2) = (hvsome (1)))
    /\ ((hread H' fork2) = (hvsome (1)) -> (hread H fork2) = (hvsome (1))));
    Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; Constructor; Auto.
Unfold getValinH in H1; Unfold getValinH; Unfold getValinR.
Inversion_clear H1; SplitAllHyps.
Constructor.
Exists x; Constructor; Auto.
Intros; Induction s'.
Unfold G2_philo; Unfold A1_philo.
SplitAllHyps; Repeat (Constructor; Auto).
Intro EQ; Rewrite EQ in H1; Inversion H1.
Intro; Absurd (hread a1 (1))=(Some nat (1)); Auto.
(* st r1 (0) r2 *)
Apply I_st with Inv_philo
 ([s:state] (EX vf2 | (getValinH s fork1)=(hvsome (2))
  /\ (getValinH s fork2)=(hvsome vf2) /\ ~vf2=(2)))
 ([s,s':state] let (H, R) = s in let (H', R') = s' in
   (hindom fork1 H') /\ (hindom fork2 H')
    /\ ((hread H' fork2) = (hvsome (2)) -> (hread H' fork1) = (hvsome (2)))
    /\ ~(hread H' fork1)=(hvsome (1))
    /\ ((hread H fork2) = (hvsome (1)) -> (hread H' fork2) = (hvsome (1)))
    /\ ((hread H' fork2) = (hvsome (1)) -> (hread H fork2) = (hvsome (1))));
    Auto.
Unfold getValinR; Unfold getValinH; Unfold Inv_philo; Intros.
Unfold plus; SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H5; Rewrite H7; Clear H5; Clear H7; Constructor; Auto.
Intros.
Cut (hread H5 (2))=(hread H (2));
 [Intro EQ; Rewrite EQ | Rewrite <- H6; Apply local_write; Auto].
Constructor; Auto.
Cut (hread H5 (1)) = (hvsome (2));
 [Intro EQ1; Rewrite EQ1 | Rewrite <- H6; Apply write_read].
Constructor; [Constructor | Exists x; Auto].
Apply indom_some with (2); Auto.
Left.
Inversion_clear H3; Auto.
Rewrite H7; Auto.
(* yield *)
Apply I_yld with Inv_philo A2_philo G2_philo; Auto.
Unfold Inv_philo; Unfold getValinH; Intros; Induction s.
SplitAllHyps; SplitAllExs; SplitAllHyps; Repeat (Constructor; Auto).
Apply indom_some with x; Auto.
Intro EQ; Rewrite EQ in H1; Inversion H1.
Unfold getValinH; Intros; Induction s; Induction s'.
Unfold Inv_philo in H; Unfold A2_philo in H0; SplitAllHyps.
SplitAllExs; SplitAllHyps.
Rewrite H8 in H3; Rewrite H8 in H4; Rewrite H6 in H5; Rewrite H6 in H7.
Inversion_clear H1; Inversion_clear H0.
Exists n0; Repeat (Constructor; Auto).
Unfold hread in H7; Rewrite H1 in H7.
Intro; Absurd (Some nat x)=(Some nat (2)); Auto.
Intro EQ; Inversion EQ; Absurd x=(2); Auto.
(* jd fb2 *)
Apply I_jd with CT2_philo Inv_philo a_fb2 g_fb2; Auto.
Unfold CT2_philo; LookupCT.
Intros; Induction s; Unfold a_fb2; SplitAllHyps; Auto.
Save.

Lemma Infer_fb2 : (Infer TSpec2_philo a_fb2 g_fb2 I_fb2).
Proof.
Unfold a_fb2; Unfold g_fb2; Unfold I_fb2.
(* yield *)
Apply I_yld with Inv_philo A2_philo G2_philo; Auto.
Unfold G2_philo; Unfold getValinH; Intros; Induction s; Unfold A1_philo.
Unfold Inv_philo in H; SplitAllHyps.
SplitAllExs; SplitAllHyps; Repeat (Constructor; Auto).
Apply indom_some with x; Auto.
Unfold getValinH; Intros; Induction s; Induction s'.
Unfold Inv_philo in H; Unfold A2_philo in H0; SplitAllHyps.
SplitAllExs; SplitAllHyps.
Rewrite H8 in H3; Rewrite H8 in H4; Rewrite H6 in H5; Rewrite H6 in H7.
Inversion_clear H1; Inversion_clear H0.
Exists n0; Repeat (Constructor; Auto).
Unfold hread in H7; Rewrite H1 in H7.
Intro; Absurd (Some nat x)=(Some nat (2)); Auto.
Intro EQ; Inversion EQ; Absurd x=(2); Auto.
(* movi r1 fork2 *)
Apply I_seq with Inv_philo
 ([s:state] (a_fb2 s) /\ (getValinR s r1)=fork2)
  G2_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; Constructor; Auto.
(* ld r2 r1 (0) *)
Apply I_ld with Inv_philo
 ([s:state] (a_fb2 s) /\ (getValinR s r1)=fork2
  /\ (hvsome (getValinR s r2))=(getValinH s fork2))
  G2_philo; Auto.
Unfold getValinR; Unfold getValinH; Unfold Inv_philo; Intros.
SplitAllHyps.
Cut (hindom (2) H).
Intro HH; Inversion HH; Rewrite H3; Unfold plus.
Exists n; Repeat (Constructor; Auto).
Inversion_clear H2.
Inversion_clear H4; [Apply indom_some with O | Apply indom_some with n]; Auto.
Unfold hread in H2; Rewrite <- H1; Auto.
(* movi r3 (0) *)
Apply I_seq with Inv_philo
 ([s:state] (a_fb2 s) /\ (getValinR s r1)=fork2
  /\ (hvsome (getValinR s r2))=(getValinH s fork2) /\ (getValinR s r3)=O)
  G2_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; Constructor; Auto.
(* bgt r2 r3 fb2 *)
Apply I_bgt with CT2_philo Inv_philo a_fb2
 ([s:state] (getValinH s fork1)=(hvsome (2))
  /\ (getValinH s fork2)=(hvsome O) /\ (getValinR s r1)=fork2)
  g_fb2 G2_philo; Auto.
Unfold CT2_philo; LookupCT.
Unfold getValinR; Unfold getValinH; Unfold a_fb2; Unfold Inv_philo; Unfold g_fb2; Unfold G2_philo; Intros.
SplitAllHyps; Constructor; Auto.
Unfold Inv_philo; Unfold a_fb2; Unfold getValinH; Unfold getValinR; Intros.
SplitAllHyps; Constructor; Auto.
SplitAllExs; SplitAllHyps; Repeat (Constructor; Auto).
Rewrite H6 in H0; Rewrite <- H4.
Replace (lookupR R r2) with O; Auto with arith.
(* movi r2 2 *)
Apply I_seq with Inv_philo
 ([s:state] (getValinH s fork1)=(hvsome (2))
  /\ (getValinH s fork2)=(hvsome O)
  /\ (getValinR s r1)=fork2 /\ (getValinR s r2)=(2))
 ([s,s':state] let (H, R) = s in let (H', R') = s' in
   (hindom fork1 H') /\ (hindom fork2 H')
    /\ ((hread H' fork2) = (hvsome (2)) -> (hread H' fork1) = (hvsome (2)))
    /\ ~(hread H' fork1)=(hvsome (1))
    /\ ~(hread H' fork2)=(hvsome (1)));
    Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo.
Unfold getValinH in H0; Unfold getValinH; Unfold getValinR in H0; Unfold getValinR; SplitAllHyps.
Repeat (Constructor; Auto).
Intros; Induction s'.
Unfold G2_philo; Unfold A1_philo; SplitAllHyps.
Repeat (Constructor; Auto).
Intro EQ; Rewrite EQ in H0; Inversion H0.
Intro; Absurd (hread a1 (1))=(Some nat (1)); Auto.
Intro EQ; Rewrite EQ in H1; Inversion H1.
Intro; Absurd (hread a1 (2))=(Some nat (1)); Auto.
(* st r1 (0) r2 *)
Apply I_st with Inv_philo
 ([s:state] (getValinH s fork1)=(hvsome (2))
  /\ (getValinH s fork2)=(hvsome (2)))
 ([s,s':state] let (H, R) = s in let (H', R') = s' in
   (hindom fork1 H') /\ (hindom fork2 H')
    /\ ((hread H' fork2) = (hvsome (2)) -> (hread H' fork1) = (hvsome (2)))
    /\ ~(hread H' fork1)=(hvsome (1))
    /\ ~(hread H' fork2)=(hvsome (1)));
    Auto.
Unfold getValinR; Unfold getValinH; Unfold Inv_philo; Intros.
Unfold plus; SplitAllHyps.
Rewrite H3; Rewrite H5; Clear H3; Clear H5; Constructor.
Apply indom_some with O; Auto.
Intros.
Cut (hread H1 (1))=(hread H (1));
 [Intro EQ; Rewrite EQ | Rewrite <- H3; Apply local_write; Auto].
Constructor; Auto.
Cut (hread H1 (2)) = (hvsome (2));
 [Intro EQ1; Rewrite EQ1 | Rewrite <- H3; Apply write_read].
Repeat (Constructor; Auto).
Rewrite H0 in EQ; Apply indom_some with (2); Auto.
(* yield *)
Apply I_yld with Inv_philo A2_philo G2_philo; Auto.
Unfold Inv_philo; Unfold getValinH; Intros; Induction s.
SplitAllHyps; Repeat (Constructor; Auto).
Apply indom_some with (2); Auto.
Intro EQ; Rewrite EQ in H; Inversion H.
Intro EQ; Rewrite EQ in H2; Inversion H2.
Unfold getValinH; Intros; Induction s; Induction s'.
Unfold Inv_philo in H; Unfold A2_philo in H0; SplitAllHyps.
Constructor; Auto.
(* movi r1 fork2 *)
Apply I_seq with Inv_philo
 ([s:state] (getValinH s fork1)=(hvsome (2))
  /\ (getValinH s fork2)=(hvsome (2)) /\ (getValinR s r1)=fork2)
  G2_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; Constructor; Auto.
(* movi r2 0 *)
Apply I_seq with Inv_philo
 ([s:state] (getValinH s fork1)=(hvsome (2))
  /\ (getValinH s fork2)=(hvsome (2)) /\ (getValinR s r1)=fork2
  /\ (getValinR s r2)=O)
  G2_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; Constructor; Auto.
(* st r1 (0) r2 *)
Apply I_st with Inv_philo
 ([s:state] (EX vf2 | (getValinH s fork1)=(hvsome (2))
  /\ (getValinH s fork2)=(hvsome vf2) /\ ~vf2=(2)))
  G2_philo; Auto.
Unfold getValinH; Unfold getValinR; Unfold Inv_philo; Unfold plus; Intros.
SplitAllHyps.
Rewrite H3; Rewrite H5; Clear H3; Clear H5; Constructor.
Apply indom_some with (2); Auto.
Intros.
Cut (hread H1 (2))=(hvsome O);
 [Intro EQ; Rewrite EQ | Rewrite <- H3; Apply write_read].
Cut (hread H1 (1))=(hread H (1));
 [Intro EQ1; Rewrite EQ1 | Rewrite <- H3; Apply local_write; Auto].
Repeat (Constructor; Auto).
Rewrite H0 in EQ1; Apply indom_some with (2); Auto.
Exists O; Auto.
Unfold G2_philo; Unfold A1_philo.
Intros; Induction s.
SplitAllHyps; Repeat (Constructor; Auto).
Intro FF; Rewrite FF in H0; Inversion H0.
Intro; Cut (hread H1 (1))=(Some nat (1)); Auto.
Intro FF; Rewrite FF in EQ1; Rewrite <- EQ1 in H0; Inversion H0.
Intro FF; Rewrite FF in H2; Inversion H2.
Intro; Cut (hread H1 (2))=(Some nat (1)); Auto.
Intro FF; Rewrite FF in EQ; Inversion EQ.
(* yield *)
Apply I_yld with Inv_philo A2_philo G2_philo; Auto.
Unfold Inv_philo; Unfold getValinH; Intros; Induction s.
Unfold G2_philo; Unfold A1_philo; SplitAllHyps; Repeat (Constructor; Auto).
Inversion_clear H2.
Apply indom_some with O; Auto.
Inversion_clear H.
Apply indom_some with n.
Rewrite <- H2; Unfold hread in H0; Auto.
Inversion_clear H2.
Intro EQ; Rewrite EQ in H0; Inversion H0.
Intro EQ; Rewrite <- EQ; Auto.
Unfold getValinH; Intros; Induction s; Induction s'.
Unfold Inv_philo in H; Unfold A2_philo in H0; SplitAllHyps.
Inversion_clear H8; SplitAllHyps.
Inversion_clear H0.
Exists n; Repeat (Constructor; Auto).
Intro EQ; Rewrite EQ in H10.
Absurd (hread a (2))=(Some nat (2)); Auto.
Rewrite H6; Intro EQ1; Inversion EQ1; Absurd x=(2); Auto.
(* movi r1 fork1 *)
Apply I_seq with Inv_philo
 ([s:state] (EX vf2 | (getValinH s fork1)=(hvsome (2))
  /\ (getValinH s fork2)=(hvsome vf2) /\ ~vf2=(2) /\ (getValinR s r1)=fork1))
  G2_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; SplitAllExs.
SplitAllHyps; Repeat (Constructor; Auto).
Exists x; Constructor; Auto.
(* movi r2 0 *)
Apply I_seq with Inv_philo
 ([s:state] (EX vf2 | (getValinH s fork1)=(hvsome (2))
  /\ (getValinH s fork2)=(hvsome vf2) /\ ~vf2=(2) /\ (getValinR s r1)=fork1
  /\ (getValinR s r2)=O))
  G2_philo; Auto.
Intros; Induction s; Induction s1.
Unfold Next in H; Inversion_clear H.
Unfold Inv_philo in H0; Unfold Inv_philo; SplitAllHyps; SplitAllExs.
SplitAllHyps; Repeat (Constructor; Auto).
Exists x; Constructor; Auto.
(* st r1 (0) r2 *)
Apply I_st with Inv_philo
 ([s:state] (EX vf1 | (EX vf2 | (getValinH s fork1)=(hvsome vf1)
  /\ (getValinH s fork2)=(hvsome vf2) /\ ~vf1=(2)/\ ~vf2=(2))))
  G2_philo; Auto.
Unfold getValinH; Unfold getValinR; Unfold Inv_philo; Unfold plus; Intros.
SplitAllHyps; SplitAllExs; SplitAllHyps.
Rewrite H5; Rewrite H7; Clear H5; Clear H7; Constructor; Auto.
Intros.
Cut (hread H5 (2))=(hread H (2));
 [Intro EQ; Rewrite EQ | Rewrite <- H6; Apply local_write; Auto].
Cut (hread H5 (1))=(hvsome O);
 [Intro EQ1; Rewrite EQ1 | Rewrite <- H6; Apply write_read].
Repeat (Constructor; Auto).
Apply indom_some with O; Auto.
Inversion_clear H3; Auto.
Rewrite H7 in H1; Rewrite H2 in H1; Inversion H1; Absurd x=(2); Auto.
Exists O; Exists x; Auto.
Unfold G2_philo; Unfold A1_philo; Intros; Induction s.
SplitAllHyps; Repeat (Constructor; Auto).
Intro FF; Rewrite FF in H2; Inversion H2.
Intro; Cut (hread H5 (1))=(Some nat (1)); Auto.
Intro FF; Rewrite FF in EQ1; Inversion EQ1.
Rewrite <- EQ; Auto.
Rewrite <- EQ; Auto.
(* yield *)
Apply I_yld with Inv_philo A2_philo G2_philo; Auto.
Unfold Inv_philo; Unfold getValinH; Intros; Induction s.
Unfold G2_philo; Unfold A1_philo; SplitAllHyps; SplitAllExs; SplitAllHyps.
Repeat (Constructor; Auto).
Apply indom_some with x0; Auto.
Intro EQ; Rewrite EQ in H1; Inversion H1; Absurd x0=(2); Auto.
Unfold getValinH; Intros; Induction s; Induction s'.
Unfold Inv_philo in H; Unfold A2_philo in H0; SplitAllHyps.
SplitAllExs; SplitAllHyps.
Rewrite H6 in H3; Rewrite H6 in H4; Rewrite H8 in H5; Rewrite H8 in H7.
Inversion_clear H1; Inversion_clear H0.
Exists n; Exists n0; Repeat (Constructor; Auto).
Unfold hread in H4; Rewrite H11 in H4.
Intro; Absurd (Some nat x)=(Some nat (2)); Auto.
Intro EQ; Inversion EQ; Absurd x=(2); Auto.
Unfold hread in H7; Rewrite H1 in H7.
Intro; Absurd (Some nat x0)=(Some nat (2)); Auto.
Intro EQ; Inversion EQ; Absurd x0=(2); Auto.
(* jd fa2 *)
Apply I_jd with CT2_philo Inv_philo a_fa2 g_fa2; Auto.
Unfold CT2_philo; LookupCT.
Intros; Induction s; Unfold a_fa2; SplitAllHyps; Auto.
Save.

Lemma WF_CH2_philo : (WF_CH TSpec2_philo C2_philo).
Proof.
Apply wf_c.
Replace (tgetCT TSpec2_philo) with CT2_philo;
 [Unfold CT2_philo; Unfold C2_philo | Auto].
Apply wf_c_cons.
Apply Infer_fa2.
Apply wf_c_cons.
Apply Infer_fb2.
Apply wf_c_emp.
Save.

Theorem WF_philo : (WFW Spec_philo p1_philo p2_philo g1_philo Prog_philo).
Proof.
Unfold Prog_philo.
Apply wf1_w with TSpec1_philo TSpec2_philo Inv_philo A2_philo G2_philo; Auto.
Apply WF_CH1_philo.
Apply WF_CH2_philo.
Unfold Inv_philo; Unfold S_philo; Constructor; Auto.
Constructor; Auto.
Apply indom_some with O; Auto.
Unfold a_fa1.
Exists O; Exists O; Auto.
Unfold S_philo; Unfold G1_philo; Intro; Unfold a_fa2.
Induction s''; Unfold A2_philo; Intro; Unfold getValinH.
SplitAllHyps.
Clear H2 H4.
Unfold hread in H3; Unfold H_philo in H3; Unfold beq_nat in H3.
Unfold hread in H6; Unfold H_philo in H6; Unfold beq_nat in H6.
Unfold hread in H1; Inversion_clear H0; Inversion_clear H.
Rewrite H2 in H1; Rewrite H2 in H3; Rewrite H0 in H1; Rewrite H0 in H6.
Exists n; Exists n0; Repeat (Constructor; Auto).
Intro; Absurd (Some nat O)=(Some nat (2)); Auto.
Intro FF; Inversion_clear FF.
Intro; Absurd (Some nat O)=(Some nat (2)); Auto.
Intro FF; Inversion_clear FF.
Unfold Inv_philo; Unfold A2_philo; Unfold a_fa2; Intros.
Induction s'; Induction s''.
Unfold getValinH in H; Unfold getValinH; Unfold hread in H; Unfold hread in H0.
Unfold hread; SplitAllHyps.
Inversion_clear H1; Inversion_clear H0.
Exists n; Exists n0; Clear H2.
Rewrite H6 in H3; Rewrite H6 in H4; Rewrite H1 in H5; Rewrite H1 in H7.
SplitAllExs.
Clear H9; SplitAllHyps.
Rewrite H0 in H3; Rewrite H0 in H4; Rewrite H2 in H5; Rewrite H2 in H7.
Rewrite H6; Rewrite H1.
Clear H; Clear H6; Clear H1; Clear H0; Clear H2; Clear a0 a.
Repeat (Constructor; Auto).
Intro; Absurd (Some nat x)=(Some nat (2)); Auto.
Intro EQ; Inversion EQ; Absurd (x=(2)); Auto.
Intro; Absurd (Some nat x0)=(Some nat (2)); Auto.
Intro EQ; Inversion EQ; Absurd (x0=(2)); Auto.
Apply I_jd with CT1_philo Inv_philo a_fa1 G1_philo; Auto.
Unfold CT1_philo.
LookupCT.
Intros; SplitAllHyps; Constructor; Auto.
Apply I_jd with CT2_philo Inv_philo a_fa2 G2_philo; Auto.
Unfold CT2_philo.
LookupCT.
Intros; SplitAllHyps; Constructor; Auto.
Save.
